using System;
using System.Drawing.Design;
using System.Web.UI;
using System.Web.UI.Design;
using System.Web.UI.WebControls;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Windows.Forms;
using System.Windows.Forms.Design;
using RepPP;

namespace RepPP.Toolkit {
    //********************************************************************************
    //
    /// <summary>
    /// Implement the base functionnality of a Rep++ control which inherit from a WEB 
    /// control and supports a card group property.
    /// Because .NET doesn't support multiple inheritance, this class must be embedded
    /// </summary>
    //
    //********************************************************************************
    public class CrdGrpControlUtil : ControlUtil {
	    /// <summary>Common interface for controls attached to a group</summary>
	    public interface ICrdGrpControl : IRepControl {
	        /// <summary>Rep++ application object</summary>
	        RepPP.Application           Application { get; }
	        /// <summary>Card Group Name</summary>
	        CrdGrpFullName              CardGroupFullName { get; set; }
	        /// <summary>Trigger an event to indicate a component property has changed</summary>
	        void                        TriggerComponentChange();
	    }
	    
	    /// <summary>CardGroup attached to the control</summary>
	    protected CardGroup             m_cgrp;
	    /// <summary>Group instance attached to the control</summary>
	    protected GroupInstance         m_ginst;
	    /// <summary>Full name of the CardGroup</summary>
	    protected CrdGrpFullName        m_crdGrpFullName;

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                           Class constuctor
        /// </summary>
        /// <param name="parent">           Control embedding this class</param>
        /// <param name="bResDispenserOnly">Must only use the Resource Dispenser to create the object</param>
        /// <param name="bNoRefCount">      true to disable reference counting for this instance</param>
        //
        //********************************************************************************
        public CrdGrpControlUtil(object parent, bool bResDispenserOnly, bool bNoRefCount) : base(parent, bResDispenserOnly, bNoRefCount) {
		    m_cgrp           = null;
		    m_ginst          = null;
		    m_crdGrpFullName = new CrdGrpFullName();
		}

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                           Class constuctor
        /// </summary>
        /// <param name="parent">           Control embedding this class</param>
        /// <param name="bResDispenserOnly">Must only use the Resource Dispenser to create the object</param>
        //
        //********************************************************************************
        public CrdGrpControlUtil(object parent, bool bResDispenserOnly) : this(parent, bResDispenserOnly, false) {
		}

        //********************************************************************************
        //
        /// <summary>
        /// CardGroupFullName:              The full name of the CardGroup attached to the control
        /// </summary>
        //
        //********************************************************************************
		public CrdGrpFullName CardGroupFullName {
		    get {
		        return(m_crdGrpFullName);
		    }
		    set {
		        m_crdGrpFullName   = value;
		        m_cgrp             = null;
		        m_ginst            = null;
            }
		}

        //********************************************************************************
        //
		/// <summary>
		/// CardGroup:                  Rep++ CardGroup attached to the control
		/// </summary>
        //
        //********************************************************************************
		public CardGroup CardGroup {
		    get {
                RepPP.CardGroup     cgrpRetVal = null;
                RepPP.Card          crd;
		        RepPP.Application   app;
		        
		        if (m_cgrp != null && m_cgrp.IsDetached) {
		            m_cgrp   = null;
		            m_ginst = null;
		        }
		        if (m_cgrp != null) {
                    cgrpRetVal = m_cgrp;
                } else {
		            app   = Application;
                    if (app != null && m_crdGrpFullName.CardName != string.Empty && m_crdGrpFullName.CardGroupName != string.Empty) {
                        crd = app.Cards[m_crdGrpFullName.CardName];
                        if (crd != null) {
                            cgrpRetVal = crd.FindCardGroup(m_crdGrpFullName.CardGroupName);
                            if (!IsDesignMode) {
                                m_cgrp = cgrpRetVal;
                            }
                        }
                    }
		        }
		        return(cgrpRetVal);
		    }
		    set {
		        if (value != m_cgrp) {
		            m_cgrp   = null;
		            m_ginst  = null;
		            if (value == null) {
		                CardGroupFullName = new CrdGrpFullName();
		            } else {
		                CardGroupFullName = new CrdGrpFullName(value.Card.Name, value.Name);
		                m_cgrp            = value;
		            }
		        }
		    }
		}

        //********************************************************************************
        //
		/// <summary>
		/// Group:                  Rep++ Group attached to the control
		/// </summary>
        //
        //********************************************************************************
		public Group Group {
		    get {
                Group       grpRetVal = null;
                Application app;
                CardGroup   cgrp;

                cgrp = CardGroup;
                if (cgrp != null) {
                    grpRetVal = cgrp.Group;
                } else {
		            app = Application;
                    if (app != null && m_crdGrpFullName.CardGroupName != string.Empty) {
                        grpRetVal = app.Groups[m_crdGrpFullName.CardGroupName];
                    }
                }
		        return(grpRetVal);
		    }
		}

        //********************************************************************************
        //
		/// <summary>
		/// GroupInstance:              Rep++ Group instance attached to the control
		/// </summary>
        //
        //********************************************************************************
		public GroupInstance GroupInstance {
		    get {
		        if (m_ginst != null && m_ginst.IsDetached) {
		            m_cgrp   = null;
		            m_ginst = null;
		        }
		        return(m_ginst);
		    }
		    set {
                CardGroup cgrp;

		        if (value != m_ginst) {
		            m_cgrp   = null;
		            m_ginst = null;
		            if (value != null) {
                        cgrp  = value.CardGroup;
                        if (cgrp != null) {
		                    CardGroupFullName = new CrdGrpFullName(cgrp.Card.Name, cgrp.Name);
                        } else {
                            CardGroupFullName = new CrdGrpFullName("", "");
                        }
		                m_cgrp  = cgrp;
		                m_ginst = value;
		            }
		        }
		    }
		}

        //********************************************************************************
        //
	    /// <summary>
	    /// Implement an editor for the CrdGrpFullName property
	    /// </summary>
        //
        //********************************************************************************
        public class CrdGrpControlEditor : System.Drawing.Design.UITypeEditor {

            //********************************************************************************
            //
            /// <summary>
            /// GetEditStyle:           Gets the style of edition
            /// </summary>
            /// <param name="context">  Context</param>
            /// <returns>
            /// Type of editor (modal)
            /// </returns>
            //
            //********************************************************************************
            public override UITypeEditorEditStyle GetEditStyle(ITypeDescriptorContext context) {
                UITypeEditorEditStyle eRetVal;
                
                if (context != null && context.Instance != null) {
                    eRetVal = UITypeEditorEditStyle.Modal;
                } else {
                    eRetVal = base.GetEditStyle(context);
                }
                return(eRetVal);
            }

            //********************************************************************************
            //
            /// <summary>
            /// EditValue:              Edit a value
            /// </summary>
            /// <param name="context">  Context</param>
            /// <param name="provider"> Provider</param>
            /// <param name="value">    Value to edit</param>
            /// <returns>
            /// New value
            /// </returns>
            //
            //********************************************************************************
            public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value) {
                object                              objRetVal = null;
                FormCgrpControlPropPage             frmCgrpControlPropPage;
                CrdGrpControlUtil.ICrdGrpControl    cgrpControl;
                IWindowsFormsEditorService          edSvc;

                if (context != null && context.Instance != null && provider != null) {
                    edSvc = provider.GetService(typeof(IWindowsFormsEditorService)) as IWindowsFormsEditorService;
                    if (edSvc != null) {
                        cgrpControl             = context.Instance as CrdGrpControlUtil.ICrdGrpControl;
#if VC8
                        if (cgrpControl == null) {
                            cgrpControl = (context.Instance as System.ComponentModel.Design.DesignerActionList).Component as CrdGrpControlUtil.ICrdGrpControl;
                        }
#endif
                        frmCgrpControlPropPage  = new FormCgrpControlPropPage(cgrpControl);
                        try {
                            edSvc.ShowDialog(frmCgrpControlPropPage);
                            objRetVal = cgrpControl.CardGroupFullName;
                        } finally {
                            frmCgrpControlPropPage.Dispose();
                        }
                    }
                }
                if (objRetVal == null) {
                    objRetVal = base.EditValue(context, provider, value);
                }
                return (objRetVal);
            }
        } // Class CrdGrpControlEditor

        //********************************************************************************
        //
	    /// <summary>
	    /// Implement an editor for the Card name property
	    /// </summary>
        //
        //********************************************************************************
        public class CardNameEditor : System.Drawing.Design.UITypeEditor {

            //********************************************************************************
            //
            /// <summary>
            /// GetEditStyle:           Gets the style of edition
            /// </summary>
            /// <param name="context">  Context</param>
            /// <returns>
            /// Type of editor (modal)
            /// </returns>
            //
            //********************************************************************************
            public override UITypeEditorEditStyle GetEditStyle(ITypeDescriptorContext context) {
                UITypeEditorEditStyle eRetVal;
                
                if (context != null && context.Instance != null) {
                    eRetVal = UITypeEditorEditStyle.DropDown;
                } else {
                    eRetVal = base.GetEditStyle(context);
                }
                return(eRetVal);
            }

            //********************************************************************************
            //
            /// <summary>
            /// EditValue:              Edit a value
            /// </summary>
            /// <param name="context">  Context</param>
            /// <param name="provider"> Provider</param>
            /// <param name="value">    Value to edit</param>
            /// <returns>
            /// New value
            /// </returns>
            //
            //********************************************************************************
            public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value) {
                object                          objRetVal = null;
                CrdGrpFullName                  cgrpFullName;
                IWindowsFormsEditorService      edSvc;
                RepPP.Application               app;
                FldControlUtil.PopupListBox     listBox;
                string                          strCrdName;

                if (context != null && context.Instance != null && provider != null) {
                    edSvc = provider.GetService(typeof(IWindowsFormsEditorService)) as IWindowsFormsEditorService;
                    app   = RepPP.Application.GetApplicationForDesign(provider);
                    if (edSvc != null && app != null) {
                        cgrpFullName            = context.Instance as CrdGrpFullName;
                        objRetVal               = cgrpFullName.CardName;
                        listBox                 = new FldControlUtil.PopupListBox(edSvc);
                        listBox.Location        = new System.Drawing.Point(0, 0);
                        listBox.Name            = "listBox";
                        listBox.Size            = new System.Drawing.Size(168, 186);
                        listBox.TabIndex        = 1;
                        listBox.BorderStyle     = System.Windows.Forms.BorderStyle.None;
                        listBox.Sorted          = true;
                        foreach (Card crd in app.Cards) {
                            if (String.Compare(crd.Name, 0, "$HIDE", 0, 5) != 0) {
                                listBox.Items.Add(crd.Name);
                            }
                        }
                        edSvc.DropDownControl(listBox);
                        if (listBox.SelectedItem != null) {
                            strCrdName = listBox.SelectedItem as string;
                            objRetVal  = strCrdName;
                            if (strCrdName != cgrpFullName.CardName) {
                                cgrpFullName.CardGroupName = "";
                            }
                        }
                    }
                }
                if (objRetVal == null) {
                    objRetVal = base.EditValue(context, provider, value);
                }
                return (objRetVal);
            }
        } // Class CardNameEditor

        //********************************************************************************
        //
	    /// <summary>
	    /// Implement an editor for the CardGroup name property
	    /// </summary>
        //
        //********************************************************************************
        public class CardGroupNameEditor : System.Drawing.Design.UITypeEditor {

            //********************************************************************************
            //
            /// <summary>
            /// GetEditStyle:           Gets the style of edition
            /// </summary>
            /// <param name="context">  Context</param>
            /// <returns>
            /// Type of editor (modal)
            /// </returns>
            //
            //********************************************************************************
            public override UITypeEditorEditStyle GetEditStyle(ITypeDescriptorContext context) {
                UITypeEditorEditStyle eRetVal;
                
                if (context != null && context.Instance != null) {
                    eRetVal = UITypeEditorEditStyle.DropDown;
                } else {
                    eRetVal = base.GetEditStyle(context);
                }
                return(eRetVal);
            }

            //********************************************************************************
            //
            /// <summary>
            /// EditValue:              Edit a value
            /// </summary>
            /// <param name="context">  Context</param>
            /// <param name="provider"> Provider</param>
            /// <param name="value">    Value to edit</param>
            /// <returns>
            /// New value
            /// </returns>
            //
            //********************************************************************************
            public override object EditValue(ITypeDescriptorContext context, IServiceProvider provider, object value) {
                object                          objRetVal = null;
                CrdGrpFullName                  cgrpFullName;
                IWindowsFormsEditorService      edSvc;
                RepPP.Application               app;
                RepPP.Card                      crd;
                FldControlUtil.PopupListBox     listBox;
                string                          strGrpName;

                if (context != null && context.Instance != null && provider != null) {
                    edSvc = provider.GetService(typeof(IWindowsFormsEditorService)) as IWindowsFormsEditorService;
                    app   = RepPP.Application.GetApplicationForDesign(provider);
                    if (edSvc != null && app != null) {
                        cgrpFullName            = context.Instance as CrdGrpFullName;
                        objRetVal               = cgrpFullName.CardGroupName;
                        listBox                 = new FldControlUtil.PopupListBox(edSvc);
                        listBox.Location        = new System.Drawing.Point(0, 0);
                        listBox.Name            = "listBox";
                        listBox.Size            = new System.Drawing.Size(168, 186);
                        listBox.TabIndex        = 1;
                        listBox.BorderStyle     = System.Windows.Forms.BorderStyle.None;
                        listBox.Sorted          = true;
                        crd                     = (cgrpFullName.CardName == string.Empty) ? null : app.Cards[cgrpFullName.CardName];
                        if (crd != null) {
                            FillCardGroupNames(crd.ChildCardGroups, listBox);
                        }
                        edSvc.DropDownControl(listBox);
                        if (listBox.SelectedItem != null) {
                            strGrpName = listBox.SelectedItem as string;
                            objRetVal  = strGrpName;
                        }   
                    }
                }
                if (objRetVal == null) {
                    objRetVal = base.EditValue(context, provider, value);
                }
                return (objRetVal);
            }

            //********************************************************************************
            //
            /// <summary>
            /// FillCardGroupNames:     Populates a list box with the names of all CardGroups in
            ///                         a card
            /// </summary>
            /// <param name="cgrps">    Card Groups Collection</param>
            /// <param name="listBox">  ListBox Control</param>
            //
            //********************************************************************************
            internal void FillCardGroupNames(CardGroups cgrps, FldControlUtil.PopupListBox listBox) {
                foreach (CardGroup cgrp in cgrps) {
                    if (String.Compare(cgrp.Name, 0, "$HIDE", 0, 5) != 0) {
                        listBox.Items.Add(cgrp.Name);
                        FillCardGroupNames(cgrp.ChildCardGroups, listBox);
                    }
                }
            }
        } // Class GroupNameEditor
    } // Class CrdGrpControlUtil
} // Namespace