using System;
using System.IO;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Xsl;

namespace RepPP.Toolkit {

	//********************************************************************************
	//
	/// <summary>
	/// Generate report from a card instance using XSL template.
    ///
    ///	(c) 2002, Consyst SQL Inc, 
	///	All rights reserved
	/// </summary>
	//
	//********************************************************************************
    public class GenXSLReport {
        private string              m_strTempDir;
        private string              m_strGuid;
        
        //******************************************************************
        //
        /// <summary>
        /// ctor:                       Class constructor
        /// </summary>
        /// <param name="strTempDir">   Temporary filename</param>
        //
        //******************************************************************
        public GenXSLReport(string strTempDir) {
            m_strTempDir = strTempDir;
            m_strGuid    = Guid.NewGuid().ToString();
        }

        //******************************************************************
        //
        /// <summary>
        /// ctor:                   Class constructor
        /// </summary>
        //
        //******************************************************************
        public GenXSLReport() : this(System.IO.Path.GetTempPath()) {
        }

        //******************************************************************
        //
        /// <summary>
        /// CreateXSLFile:                  Create a XSL file using 
        /// </summary>
        /// <param name="strCrdName">       Name of the card</param>
        /// <param name="strXSLDocument">   XLST Template source</param>
        /// <returns>
        /// XML Filename
        /// </returns>
        //
        //******************************************************************
        private string CreateXSLFile(string strCrdName, string strXSLDocument) {
            const string    cstrFILENAME    = "xlsReport{0}{1}.xsl";
            StreamWriter    streamWriter    = null;
            string          strFilename;

            strFilename  = System.IO.Path.Combine(m_strTempDir, 
                                                  string.Format(cstrFILENAME, m_strGuid, strCrdName));
            using(streamWriter = File.CreateText(strFilename)) {
                streamWriter.Write(strXSLDocument);
            } 
            return(strFilename);
        }

        //******************************************************************
        //
        /// <summary>
        /// GenerateReport:                 Generates a report in a browser 
        /// </summary>
        /// <param name="cinst">            Card Instance to generate Report</param>
        /// <param name="strXSLDocument">   XLST Template source</param>
        /// <param name="bHaveChild">       Generate XLS for Child group instance</param>
        /// <returns>
        /// XML Filename
        /// </returns>
        //
        //******************************************************************
        public string GenerateReport(RepPP.CardInstance cinst, 
                                     string             strXSLDocument, 
                                     bool               bHaveChild) {
            string          strRetVal       = null;

            CreateXSLFile(cinst.Card.Name, strXSLDocument);
            strRetVal = SerializeToXml(cinst, bHaveChild);
            return(strRetVal);
        }

        //******************************************************************
        //
        /// <summary>
        /// GenerateHTMLReport:             Generate HTML Report
        /// </summary>
        /// <param name="cinst">            Card Instance to generate Report</param>
        /// <param name="strXSLDocument">   XLST Template source</param>
        /// <param name="bHaveChild">       Generate XLS for Child group instance</param>
        /// <returns>
        /// HTML Filepath
        /// </returns>
        //
        //******************************************************************
        public string GenerateHTMLReport(RepPP.CardInstance cinst, 
                                         string             strXSLDocument, 
                                         bool               bHaveChild) {
            const string    cstrXSLFILENAME     = "xlsReport{0}{1}.xsl";
            const string    cstrHTMLFILENAME    = "report{0}{1}.html";            
#if (VC70 || VC71)
			XslTransform    xlsTransform;
#else
			XslCompiledTransform xlsTransform;
#endif
            XmlUrlResolver  resolver;
            string          strXslFilename;
            string          strXmlFileName;
            string          strRetVal           = null;

#if (VC70 || VC71)
            xlsTransform         = new XslTransform();
#else
			xlsTransform		 = new XslCompiledTransform();
#endif
            resolver             = new XmlUrlResolver();            
            strXslFilename       = System.IO.Path.Combine(m_strTempDir, 
                                                          string.Format(cstrXSLFILENAME, m_strGuid, cinst.Card.Name));
            strRetVal            = System.IO.Path.Combine(m_strTempDir, 
                                                          string.Format(cstrHTMLFILENAME, m_strGuid, cinst.Card.Name));
            strXmlFileName       = GenerateReport(cinst, strXSLDocument, bHaveChild);  
            resolver.Credentials = System.Net.CredentialCache.DefaultCredentials;
#if (VC70 || VC71)
            xlsTransform.Load(strXslFilename, resolver);
			xlsTransform.Transform(strXmlFileName, strRetVal, resolver);
#else
			XmlWriter			xmlWriter;
            XmlReader           xmlReader;
			XmlReaderSettings   settings = new XmlReaderSettings();
      
            settings.ProhibitDtd  = false;
            using (xmlReader = XmlReader.Create(strXmlFileName, settings)) {
                xlsTransform.Load(strXslFilename, XsltSettings.Default, resolver);
                using (FileStream stream1 = new FileStream(strRetVal, FileMode.Create, FileAccess.Write)) {
                    using (xmlWriter = XmlWriter.Create(stream1)) {
                        xlsTransform.Transform(xmlReader, null, xmlWriter, resolver);
                    }
                }
            }
#endif
			return(strRetVal);
        }

        //******************************************************************
        //
        /// <summary>
        /// GenerateReport:                 Generate XLS Report without Child
        /// </summary>
        /// <param name="cinst">            Card Instance to generate Report</param>
        /// <param name="strXSLDocument">   XLST Template source</param>
        /// <returns>
        /// XSL Filepath
        /// </returns>
        //
        //******************************************************************
        public string GenerateReport(RepPP.CardInstance cinst, string strXSLDocument) {
            return(GenerateReport(cinst, strXSLDocument, false));
        }

        //******************************************************************
        //
        /// <summary>
        /// GetFieldDescription:            Get Description from ChoiceList Field 
        /// </summary>
        /// <param name="fld">              Field to get Description</param>
        /// <returns>
        /// Field Description
        /// </returns>
        //
        //******************************************************************
        private string GetFieldDescription(RepPP.Field fld) {
            string[]  arrValue;
            string    strRetVal = "";
            int       iIndex;

            if (fld.IsChoiceAvailable) {
                switch(fld.Type) {
                case RepPP.FieldType.sdFieldSimpleAlphaChoice :
                case RepPP.FieldType.sdFieldSimpleNumChoice   :
                    iIndex = fld.ChoiceList.FindValue(fld.Value);
                    if (iIndex != -1) {
                        strRetVal = fld.ChoiceList.GetDesc(iIndex);
                    }
                    break;
                case RepPP.FieldType.sdFieldMultiAlphaChoice :
                case RepPP.FieldType.sdFieldMultiNumChoice   :
                    arrValue = fld.ExtValue.Split(',');
                    foreach(string strValue in arrValue) {
                        iIndex = fld.ChoiceList.FindCode(strValue);
                        if (iIndex != -1) {
                            strRetVal += (strRetVal != string.Empty) ? "," + fld.ChoiceList.GetDesc(iIndex) :
                                                                       fld.ChoiceList.GetDesc(iIndex); 
                        }
                    }
                    break;
                }
            }
            return(strRetVal);
        }

        //******************************************************************
        //
        /// <summary>
        /// GenerateXML:            Generate XML from a group Instance                
        /// </summary>
        /// <param name="xmlFather">XML DOM</param>
        /// <param name="ginst">    Current Group Instance</param>
        /// <param name="bChild">   Generate XML for Child group instance</param>
        //
        //******************************************************************
        private void GenerateXML(XmlNode xmlFather, RepPP.GroupInstance ginst, bool bChild) {
            XmlDocument     xmlDoc;
            XmlNode         xmlGinst;
            XmlNode         xmlLines;
            XmlNode         xmlLine;
            XmlNode         xmlFields;
            XmlNode         xmlNode;
            XmlAttribute    xmlNewAttr;
            string          strFldDesc;
            int             iLineCount;            
            
            xmlDoc    = (xmlFather is XmlDocument) ? xmlFather as XmlDocument : xmlFather.OwnerDocument;
            xmlGinst  = xmlFather.AppendChild(xmlDoc.CreateElement(ginst.Group.Name));
            xmlFields = xmlGinst.AppendChild(xmlDoc.CreateElement("FIELDS"));
            xmlLines  = xmlGinst.AppendChild(xmlDoc.CreateElement("LINES"));
   			foreach (RepPP.Field fld in ginst.Group.Fields) {
                xmlNode           = xmlFields.AppendChild(xmlDoc.CreateElement(fld.Name));
                xmlNode.InnerText = fld.Prompt;
			}
			ginst.Group.SaveActiveGroupInstance();
            ginst.SaveLinePos();
            ginst.Select(false);
            try {
                iLineCount = ginst.LineCount;
                for (int i=0; i < iLineCount; i++) {
                    ginst.SelectLine(i, false, false);
                    if (ginst.LineState != RepPP.LineState.sdLineDeleted) {
                        xmlLine = xmlLines.AppendChild(xmlDoc.CreateElement("LINE"));
   			            foreach (RepPP.Field fld in ginst.Group.Fields) {
                            xmlNode           = xmlLine.AppendChild(xmlDoc.CreateElement(fld.Name));
                            xmlNode.InnerText = fld.ExtValue;
                            xmlNewAttr        = xmlNode.Attributes.Append(xmlDoc.CreateAttribute("INTVALUE"));
                            xmlNewAttr.Value  = fld.Value;                    
                            if (fld.IsChoiceAvailable) {
                                strFldDesc = GetFieldDescription(fld);
                                if (strFldDesc != string.Empty) {
                                    xmlNewAttr       = xmlNode.Attributes.Append(xmlDoc.CreateAttribute("DESCRIPTION"));
                                    xmlNewAttr.Value = strFldDesc;                    
                                }
                            }
                        }
                        if (bChild) {
                            foreach(RepPP.GroupInstance ginstChild in ginst.ChildGroupInstances) {
                                GenerateXML(xmlLine, ginstChild, true);
                            }
                        }
                    }
                }
            } finally {
                ginst.RestoreLinePos(false);
                ginst.Group.RestoreActiveGroupInstance(false);
            }
        }

        //******************************************************************
        //
        /// <summary>
        /// SerializeToXml:         Save Card Instance to XML File
        /// </summary>
        /// <param name="cinst">    Card Instance to save</param>
        /// <param name="bChild">   Build XML Child Group Instance</param>
        /// <returns>
        /// XML Filepath
        /// </returns>
        //
        //******************************************************************
        public string SerializeToXml(RepPP.CardInstance cinst, bool bChild) {
            const string    cstrENCODING     = "UTF-8";
            const string    cstrFILENAME_XML = "xlsReport{0}{1}.xml";
            const string    cstrPROCESSOR    = @"type=""text/xsl"" href=""xlsReport{0}{1}.xsl""";
            XmlDocument     xmlDoc;
            XmlTextWriter   xmlWriter        = null;
            string          strRetVal        = null;

            xmlDoc               = new XmlDocument();
            strRetVal            = System.IO.Path.Combine(m_strTempDir, 
                                                          string.Format(cstrFILENAME_XML, m_strGuid, cinst.Card.Name)); 
            xmlWriter = new XmlTextWriter(strRetVal, System.Text.Encoding.GetEncoding(cstrENCODING));
            try {
                xmlWriter.Formatting = Formatting.Indented;
                xmlDoc.AppendChild(xmlDoc.CreateXmlDeclaration("1.0", cstrENCODING, null));
                xmlDoc.AppendChild(xmlDoc.CreateProcessingInstruction("xml-stylesheet", 
                                                                      string.Format(cstrPROCESSOR, m_strGuid, cinst.Card.Name)));
                GenerateXML(xmlDoc, cinst.RootGroupInstance, bChild);
                xmlDoc.Save(xmlWriter);
            } finally {
                xmlWriter.Close();
            }
            return(strRetVal);
        }
    } // End class
} // End namespace
