using System;
using System.Collections;

namespace RepPP.Toolkit {
	/// <summary>
	/// Handle the historic for a transaction
	/// </summary>
	public abstract class HistoricBase : IDisposable {
        /// <summary>Type of operation to log</summary>
        public enum LogTypeE {
            /// <summary>No log</summary>
            None     =  0,
            /// <summary>Log only when row is inserted</summary>
            AtInsert =  1,
            /// <summary>Log only when row is updated</summary>
            AtUpdate =  2,
            /// <summary>Log only when row is deleted</summary>
            AtDelete =  4,
            /// <summary>Log when row is inserted, updated or deleted</summary>
            AtAll    =  7
        };
        
        /// <summary>Counter to give unique name to each created group</summary>
        static private int              m_iCount = 0;
        /// <summary>Rep++ Application object</summary>
        protected RepPP.Application     m_app;
        /// <summary>Rep++ database connection to use</summary>
        protected RepPP.Connection      m_conn;
        /// <summary>Card group for which the historic is done</summary>
        protected RepPP.CardGroup       m_cgrpTable;
        /// <summary>Card which can be used to read the historic</summary>
        protected RepPP.Card            m_crdHistoric;
        /// <summary>Card instance which can be used to read the historic</summary>
        protected RepPP.CardInstance    m_cinstHistoric;
        /// <summary>Group pointing to the historic table</summary>
        protected RepPP.Group           m_grpHistoric;
        /// <summary>Group instance for the historic table</summary>
        protected RepPP.GroupInstance   m_ginstHistoric;
        /// <summary>SQL command object use to insert the line in the log</summary>
        protected RepPP.SqlCommand      m_sqlInsert;
        /// <summary>SQL command to insert into the historic table</summary>
        private string                  m_strInsSqlCmd;
        /// <summary>Type of log</summary>
        protected LogTypeE              m_eLogType;

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                               Class constructor
        /// </summary>
        /// <param name="conn">                 Database connection to use (null for DataConnection)</param>
        /// <param name="strHistTablePrefix">   Prefix to add to the table name to build the historic table name</param>
        /// <param name="strHistTableSuffix">   Suffix to add to the table name to build the historic table name</param>
        /// <param name="eLogType">             Type of event to log. More than one option can be specified</param>
        /// <param name="cgrpTable">            Card Group pointing to the table to check for historic</param>
        /// <param name="grpInfo">              Group containing the additional field to put in the historic</param>
		/// <param name="arrFldName">           List of fields coming from the base group which must be
		///                                     included in the historic group. null for all</param>
		/// <param name="bInfoBefore">          true to put the information before the other fields</param>
        //
        //********************************************************************************
		protected HistoricBase(RepPP.Connection  conn,
		                       string            strHistTablePrefix,
		                       string            strHistTableSuffix,
		                       LogTypeE          eLogType,
		                       RepPP.CardGroup   cgrpTable,
		                       RepPP.Group       grpInfo,
		                       string[]          arrFldName,
		                       bool              bInfoBefore) {
	        ArrayList   arrFldName2;
	        string      strNewTableName;
	        string      strDummy = "";
	        
	        m_app           = cgrpTable.Application;
	        m_conn          = (conn == null) ? m_app.DataConnection : conn;
	        arrFldName2     = (arrFldName == null) ? null : new ArrayList(arrFldName);
	        strNewTableName = null;
	        foreach (RepPP.Field fld in cgrpTable.Group.Fields) {
	            if (fld.IsPrimaryKey) {
	                strNewTableName = fld.TableName;
	                break;
	            }
	        }
	        if (strNewTableName == null) {
	            throw new System.ApplicationException("The group " + cgrpTable.Group.Name + " must have a primary key and an attached table");
	        }
	        strNewTableName = BuildHistTableName(strNewTableName, strHistTablePrefix, strHistTableSuffix);
	        m_eLogType      = eLogType;
	        m_cgrpTable     = cgrpTable;
	        m_grpHistoric   = BuildNewGroup("$HIST" + (m_iCount++).ToString(),
	                                        strNewTableName,
	                                        cgrpTable,
	                                        grpInfo,
	                                        arrFldName2,
	                                        bInfoBefore);
	        m_crdHistoric   = m_app.Cards.Add(m_grpHistoric.Name);
	        m_crdHistoric.ChildCardGroups.Add(m_grpHistoric,
	                                          RepPP.GroupType.sdHorizontal,
	                                          RepPP.GroupInstFlag.sdNormal,
	                                          false,
	                                          0,
	                                          0,
	                                          m_conn,
	                                          RepPP.SqlOptimization.sdOptFull);
	        m_crdHistoric.BuildSqlCommand();
	        m_cinstHistoric = m_crdHistoric.CardInstances.Add();
            m_ginstHistoric = m_grpHistoric.GroupInstances.Create(RepPP.GroupType.sdVertical);
            m_grpHistoric.BuildSqlCommand("", ref strDummy, ref m_strInsSqlCmd, ref strDummy, ref strDummy);
            if ((eLogType & LogTypeE.AtInsert) != 0) {
                m_app.DbAfterInsLine += new RepPP.DbAfterInsLineEventHandler(m_app_DbAfterInsLine);
            }
            if ((eLogType & LogTypeE.AtUpdate) != 0) {
                m_app.DbAfterUpdLine += new RepPP.DbAfterUpdLineEventHandler(m_app_DbAfterUpdLine);
            }
            if ((eLogType & LogTypeE.AtDelete) != 0) {
                m_app.DbAfterDelLine += new RepPP.DbAfterDelLineEventHandler(m_app_DbAfterDelLine);
            }
		}

        //********************************************************************************
        //
		/// <summary>
		/// Dispose:                Dispose the object
		/// </summary>
        //
        //********************************************************************************
		public virtual void Dispose() {
            m_app.DbAfterInsLine -= new RepPP.DbAfterInsLineEventHandler(m_app_DbAfterInsLine);
            m_app.DbAfterUpdLine -= new RepPP.DbAfterUpdLineEventHandler(m_app_DbAfterUpdLine);
            m_app.DbAfterDelLine -= new RepPP.DbAfterDelLineEventHandler(m_app_DbAfterDelLine);
		    if (m_cinstHistoric != null) {
		        m_cinstHistoric.Delete();
		        m_cinstHistoric = null;
		    }
		    if (m_crdHistoric != null) {
		        m_crdHistoric.Delete();
		        m_crdHistoric = null;
		    }
		    if (m_sqlInsert != null) {
		        m_sqlInsert.Dispose();
		        m_sqlInsert = null;
		    }
		    if (m_ginstHistoric != null) {
		        m_ginstHistoric.Delete();
		        m_ginstHistoric = null;
		    }
		    if (m_grpHistoric != null) {
		        m_grpHistoric.Delete();
		        m_grpHistoric = null;
		    }
		}

        //********************************************************************************
        //
		/// <summary>
		/// HistoricCardInstance:   Card instance which can be used to read the historic
		/// </summary>
        //
        //********************************************************************************
		public RepPP.CardInstance HistoricCardInstance {
		    get {
		        return(m_cinstHistoric);
		    }
		}

        //********************************************************************************
        //
		/// <summary>
		/// InsertSqlCmd:           SQL command use to insert in the historic
		/// </summary>
        //
        //********************************************************************************
		public string InsertSqlCmd {
		    get {
		        return(m_strInsSqlCmd);
		    }
		    set {
		        m_strInsSqlCmd = value;
		        if (m_sqlInsert != null) {
		            m_sqlInsert.Dispose();
		            m_sqlInsert = null;
		        }
		    }
		}

        //********************************************************************************
        //
		/// <summary>
		/// BuildHistTableName:                 Routine use to build the historic table name from the base
		///                                     table name. This method can be override to specify a new way to
		///                                     get the historic table name
		/// </summary>
		/// <param name="strTableName">         Base table name</param>
		/// <param name="strHistTablePrefix">   Prefix to add</param>
		/// <param name="strHistTableSuffix">   Suffix to add</param>
		/// <returns>
		/// Historic table name
		/// </returns>
        //
        //********************************************************************************
		protected virtual string BuildHistTableName(string  strTableName,
		                                            string  strHistTablePrefix,
		                                            string  strHistTableSuffix) {
		    string  strRetVal = "";
		    
		    if (strHistTablePrefix != null) {
		        strRetVal += strHistTablePrefix;
		    }
		    strRetVal += strTableName;
		    if (strHistTableSuffix != null) {
		        strRetVal += strHistTableSuffix;
		    }
		    return(strRetVal);
		}

        //********************************************************************************
        //
		/// <summary>
		/// AddField:                   Add a field to the group using another field as template
		/// </summary>
		/// <param name="grp">          Group</param>
		/// <param name="fldSrc">       Source field to use as template</param>
		/// <param name="strTableName"> Name of the table attached to the new field</param>
		/// <param name="iLogPos">      Position of the field in the group</param>
        //
        //********************************************************************************
        private void AddField(RepPP.Group grp, RepPP.Field fldSrc, string strTableName, int iLogPos) {
		    RepPP.FieldOptions      eFldOptions;
		    RepPP.Field             fldNew;
		    
		    eFldOptions = (fldSrc.IsPrimaryKey) ? RepPP.FieldOptions.sdFieldPK : (RepPP.FieldOptions)0;
		    if (fldSrc.IsPassword) {
		        eFldOptions |= RepPP.FieldOptions.sdFieldNoEcho;
		    }
		    fldNew   = grp.Fields.Add(fldSrc.Name,
		                              iLogPos,
		                              strTableName,
		                              fldSrc.ColName,
		                              fldSrc.Type,
		                              eFldOptions,
		                              fldSrc.MinSize,
                                      fldSrc.MaxSize,
		                              fldSrc.Precision);
		    if (fldSrc.IsChoiceAvailable && fldSrc.ChoiceList.List != null) {
		        fldNew.ChoiceList.List = fldSrc.ChoiceList.List;
		    }
		    fldNew.InputMask  = fldSrc.InputMask;
		    fldNew.Prompt     = fldSrc.Prompt;
		    fldNew.OutputMask = fldSrc.OutputMask;
		    fldNew.IsVisible  = fldSrc.IsVisible;
		    
        }

        //********************************************************************************
        //
        /// <summary>
        /// AddFatherPKField:               Add the primary key fields of the ancesters
        /// </summary>
        /// <param name="grp">              Group to add the field to</param>
		/// <param name="strNewTableName">  Historic table name</param>
        /// <param name="cgrpFather">       Father card group</param>
        /// <param name="iLogPos">          Actual position</param>
        //
        //********************************************************************************
        private void AddFatherPKField(RepPP.Group grp, string strNewTableName, RepPP.CardGroup cgrpFather, ref int iLogPos) {
            if (cgrpFather != null) {
                AddFatherPKField(grp, strNewTableName, cgrpFather.Parent as RepPP.CardGroup, ref iLogPos);
                foreach (RepPP.Field fld in cgrpFather.Group.Fields) {
                    if (fld.IsPrimaryKey) {
                        AddField(grp, fld, strNewTableName, iLogPos++);                        
                    }
                }
            }
        }

        //********************************************************************************
        //
		/// <summary>
		/// BuildNewGroup:                  Build a new group which represent the record to
		///                                 insert in the historic table
		/// </summary>
		/// <param name="strNewGrpName">    Name of the new group</param>
		/// <param name="strNewTableName">  Historic table name</param>
		/// <param name="cgrpTable">        Base card group</param>
		/// <param name="grpInfo">          Group containing the fields to the historic group</param>
		/// <param name="arrFldName">       List of fields coming from the base group which must be
		///                                 included in the historic group. null for all</param>
		/// <param name="bInfoBefore">      true to put the information before the other fields</param>
		/// <returns>
		/// New group
		/// </returns>
        //
        //********************************************************************************
		private RepPP.Group BuildNewGroup(string            strNewGrpName,
		                                  string            strNewTableName,
		                                  RepPP.CardGroup   cgrpTable,
		                                  RepPP.Group       grpInfo,
		                                  ArrayList         arrFldName,
		                                  bool              bInfoBefore) {
		    RepPP.Group             grpRetVal;
		    int                     iLogPos = 1;
		    
		    grpRetVal = m_app.Groups.Add(strNewGrpName, "", true);
		    if (bInfoBefore) {
		        iLogPos = 500;
		    }
		    AddFatherPKField(grpRetVal, strNewTableName, cgrpTable.Parent as RepPP.CardGroup, ref iLogPos);
		    foreach (RepPP.Field fld in cgrpTable.Group.Fields) {
		        if (fld.IsDatabaseLinked) {
		            if (arrFldName == null ||
		                arrFldName.IndexOf(fld.Name) != -1) {
		                AddField(grpRetVal, fld, strNewTableName, iLogPos++);
		            }
		        }
		    }
		    if (bInfoBefore) {
		        iLogPos = 1;
		    }
		    foreach (RepPP.Field fld in grpInfo.Fields) {
		        AddField(grpRetVal, fld, strNewTableName, iLogPos++);
		    }
		    return(grpRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// FillLogRecord:                  Fill the extra field in the historic group. This method
        ///                                 must be overriden.
        /// </summary>
        /// <param name="eType">            Type of log (insert, update, delete)</param>
        /// <param name="ginstSrc">         Group instance containing the value of the base table which is inserted, updated or deleted</param>
        /// <param name="ginstLogRecord">   Historic group instance that you must complete</param>
        /// <returns>
        /// true if ok, false if error.
        /// </returns>
        //
        //********************************************************************************
        protected abstract bool FillLogRecord(LogTypeE              eType,
                                              RepPP.GroupInstance   ginstSrc,
                                              RepPP.GroupInstance   ginstLogRecord);

        //********************************************************************************
        //
        /// <summary>
        /// InsertLogRecord:                Insert a log record in the database
        /// </summary>
        /// <param name="eType">            Type of log record (insert, update, delete)</param>
        /// <param name="ginstSrc">         Group instance containing the value of the base table</param>
        /// <param name="ginstLogRecord">   Historic group instance that you must complete</param>
        /// <returns>
        /// true if success, false if error.
        /// </returns>
        //
        //********************************************************************************
        protected virtual bool InsertLogRecord(LogTypeE              eType,
                                               RepPP.GroupInstance   ginstSrc,
                                               RepPP.GroupInstance   ginstLogRecord) {
            bool                    bRetVal;
            RepPP.GroupInstance     ginstFather;
            
            m_ginstHistoric.Group.SaveActiveGroupInstance();
            try {
                m_ginstHistoric.Select(false, false);
                m_ginstHistoric.Empty();
                ginstFather = ginstSrc.FatherGroupInstance;
                while (ginstFather != null) {
                    m_grpHistoric.CopyFieldsFrom(ginstFather.Group, RepPP.CopyFieldOpt.sdSetValid);
                    ginstFather = ginstFather.FatherGroupInstance;
                }
                m_grpHistoric.CopyFieldsFrom(ginstSrc.Group, RepPP.CopyFieldOpt.sdSetValid);
                bRetVal = FillLogRecord(eType, ginstSrc, ginstLogRecord);
                if (bRetVal) {
                    if (m_sqlInsert == null) {
                        m_sqlInsert = m_conn.SqlCommands.Open(m_strInsSqlCmd, RepPP.SqlOptimization.sdOptCursor);
                    }
                    if (m_sqlInsert.Execute() != 0) {
                        m_app.Errors.Add(0,
                                        RepPP.ErrorLevel.sdErrValidation,
                                        m_conn.ErrorMessage,
                                        m_conn.Name,
                                        RepPP.ObjType.sdGroupInst,
                                        ginstLogRecord);
                        bRetVal = false;
                        m_conn.Rollback();
                    }
                }
            } finally {
                m_ginstHistoric.Group.RestoreActiveGroupInstance(false);
            }
            return(bRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// m_app_DbAfterInsLine:   Called when a line is inserted
        /// </summary>
        /// <param name="sender">   Sender object</param>
        /// <param name="e">        Event parameter</param>
        //
        //********************************************************************************
        private void m_app_DbAfterInsLine(object sender, RepPP.GroupInstanceEventArgs e) {
            if (e.GroupInstance.Group == m_cgrpTable.Group) {
                if (!InsertLogRecord(LogTypeE.AtInsert,
                                     e.GroupInstance,
                                     m_ginstHistoric)) {
                    e.EventState = RepPP.EventState.sdReject;
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// m_app_DbAfterUpdLine:   Called when a line is updated
        /// </summary>
        /// <param name="sender">   Sender object</param>
        /// <param name="e">        Event parameter</param>
        //
        //********************************************************************************
        private void m_app_DbAfterUpdLine(object sender, RepPP.GroupInstanceEventArgs e) {
            if (e.GroupInstance.Group == m_cgrpTable.Group) {
                if (!InsertLogRecord(LogTypeE.AtUpdate,
                                     e.GroupInstance,
                                     m_ginstHistoric)) {
                    e.EventState = RepPP.EventState.sdReject;
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// m_app_DbAfterDelLine:   Called when a line is deleted
        /// </summary>
        /// <param name="sender">   Sender object</param>
        /// <param name="e">        Event parameter</param>
        //
        //********************************************************************************
        private void m_app_DbAfterDelLine(object sender, RepPP.GroupInstanceEventArgs e) {
            if (e.GroupInstance.Group == m_cgrpTable.Group) {
                if (!InsertLogRecord(LogTypeE.AtDelete,
                                     e.GroupInstance,
                                     m_ginstHistoric)) {
                    e.EventState = RepPP.EventState.sdReject;
                }
            }
        }
    }
}
