using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Xml;

namespace RepPP.Toolkit {
	//******************************************************************
	//
	/// <summary>
	/// Implement a base class for filter of a selection buffer
	/// </summary>
	//
	//******************************************************************
	public abstract class SBFilterBase : ISBFilter {
	    /// <summary>Specifies if we must use a AND or a OR between condition in the WHERE clause</summary>
	    public enum AndOrE {
	        /// <summary>Use AND expression</summary>
	        AND =   1,
	        /// <summary>Use OR expression</summary>
	        OR  =   2
	    };
	    
	    /// <summary>Counter to create unique group name</summary>
	    private static int                      m_iGlobCounter = 0;
	    /// <summary>Value that represents an empty choice in a choice list</summary>
        private const string                    m_strNoSelection = "*";
	    /// <summary>Rep++ Application object</summary>
	    protected RepPP.IApplication            m_app;
	    /// <summary>Temp group instance containing the filter values</summary>
	    protected RepPP.GroupInstance           m_ginstParam;
	    /// <summary>Group use as a template</summary>
	    protected RepPP.Group                   m_grpTemplate;
	    /// <summary>Group containing the filter values</summary>
	    protected RepPP.Group                   m_grpParam;
	    /// <summary>Actual WHERE clause</summary>
	    protected string                        m_strWhereClause;
	    /// <summary>Human readable WHERE clause</summary>
	    protected string                        m_strFilterClause;
        /// <summary>wild any char for like clause</summary>
        protected char                          m_chrWildCharAny;
        /// <summary>wild one char for like clause</summary>
        protected char                          m_chrWildCharOne;
        /// <summary>String shown for BETWEEN clause</summary>
        protected string                        m_strBetween;
        /// <summary>Array of generated fields</summary>
        protected Field[]                       m_arrFld;
        /// <summary>Array of generated BETWEEN fields</summary>
        protected Field[]                       m_arrBetweenFld;
        /// <summary>Indicateds whether or not to show the user flag field on the filter</summary>
        protected bool                          m_bShowUserFlagField;

	    //******************************************************************
	    //
	    /// <summary>
	    /// ctor:                               Class constructor
	    /// </summary>
		/// <param name="grpParam">             Group containing the field use as filter</param>
		/// <param name="bShowUserFlagField">   True to add user flag fields to the filter group</param>
	    //
	    //******************************************************************
		public SBFilterBase(RepPP.Group grpParam, bool bShowUserFlagField) {
            m_arrFld                = null;
            m_arrBetweenFld         = null;
		    m_app                   = grpParam.Application;
		    m_ginstParam            = null;
            m_bShowUserFlagField    = bShowUserFlagField;
		    TemplateGroup           = grpParam;
		    m_strWhereClause        = "";
		    m_strFilterClause       = "";
            m_chrWildCharAny        = '%';
            m_chrWildCharOne        = '_';
            m_strBetween            = "between";
		}

	    //******************************************************************
	    //
	    /// <summary>
	    /// ctor:                       Class constructor
	    /// </summary>
		/// <param name="grpParam">     Group containing the field use as filter</param>
	    //
	    //******************************************************************
		public SBFilterBase(RepPP.Group grpParam) : this (grpParam, true) {
		}

	    //******************************************************************
	    //
	    /// <summary>
	    /// Dispose:                    Dispose the class
	    /// </summary>
	    //
	    //******************************************************************
		public virtual void Dispose() {
		    TemplateGroup = null;
		}

        //********************************************************************************
        //
		/// <summary>
		/// AddField:                   Add a field to the group using another field as template
		/// </summary>
		/// <param name="grp">          Group</param>
		/// <param name="fldSrc">       Source field to use as template</param>
		/// <param name="strFldName">   Name of the field</param>
		/// <param name="bIsVisible">   Is the field visible?</param>
		/// <param name="iLogPos">      Position of the field in the group</param>
        //
        //********************************************************************************
        protected virtual Field AddField(RepPP.Group grp,
                                         RepPP.Field fldSrc, 
                                         string      strFldName, 
                                         bool        bIsVisible,
                                         int         iLogPos) {
		    RepPP.FieldOptions      eFldOptions;
		    RepPP.Field             fldNew;
		    
		    eFldOptions        = (fldSrc.IsPassword) ? RepPP.FieldOptions.sdFieldNoEcho : (RepPP.FieldOptions)0;
		    fldNew             = grp.Fields.Add(strFldName,
		                                        iLogPos,
		                                        fldSrc.TableName,
		                                        fldSrc.ColName,
		                                        fldSrc.Type,
		                                        eFldOptions,
		                                        fldSrc.MinSize,
                                                fldSrc.MaxSize,
		                                        fldSrc.Precision);
            fldNew.ControlType = fldSrc.ControlType;
            // Exception in case of a dynamic choice list
		    if (fldSrc.IsChoiceAvailable /*&& fldSrc.ChoiceListAtomName == String.Empty*/ && fldSrc.ChoiceList.List != null) {
		        fldNew.ChoiceList.List = fldSrc.ChoiceList.List;
                fldNew.ChoiceList.AddChoice(m_strNoSelection, 
                                            m_strNoSelection, 
                                            m_strNoSelection, 
                                            0);
		    }
		    fldNew.InputMask    = fldSrc.InputMask;
		    fldNew.Prompt       = fldSrc.Prompt;
		    fldNew.OutputMask   = fldSrc.OutputMask;
		    fldNew.IsVisible    = bIsVisible;
            fldNew.IsUserFlagOn = fldSrc.IsUserFlagOn;
		    return(fldNew);
        }

        //********************************************************************************
        //
		/// <summary>
		/// BuildNewGroup:                  Build a new group which represent the record to
		///                                 insert in the historic table
		/// </summary>
		/// <param name="strNewGrpName">    Name of the new group</param>
		/// <param name="grpSrc">           Base group</param>
		/// <returns>
		/// New group
		/// </returns>
        //
        //********************************************************************************
		protected virtual RepPP.Group BuildNewGroup(string            strNewGrpName,
		                                            RepPP.Group       grpSrc) {
		    RepPP.Group             grpRetVal;
		    int                     iLogPos = 1;
		    int                     iCount = 0;
		    ArrayList               arrFld;
		    ArrayList               arrBetweenFld;
		    
		    arrFld          = new ArrayList(grpSrc.Fields.Count);
		    arrBetweenFld   = new ArrayList(grpSrc.Fields.Count);
		    grpRetVal       = m_app.Groups.Add(strNewGrpName, "", true);
		    foreach (RepPP.Field fld in grpSrc.Fields) {
		        if (fld.IsDatabaseLinked                    && 
                    fld.Type != RepPP.FieldType.sdFieldMemo &&
                    (!fld.IsUserFlagOn || ShowUserFlagField)) {
                    arrFld.Add(AddField(grpRetVal, fld, fld.Name, true, iLogPos++));
                    arrBetweenFld.Add(AddField(grpRetVal, fld, "_" + (iCount++).ToString(), false, iLogPos++));
		        }
		    }
		    m_arrFld        = new Field[arrFld.Count];
		    m_arrBetweenFld = new Field[arrBetweenFld.Count];
		    arrFld.CopyTo(m_arrFld);
		    arrBetweenFld.CopyTo(m_arrBetweenFld);
		    return(grpRetVal);
        }

	    //******************************************************************
	    //
		/// <summary>
		/// TemplateGroup:          Group used as a template
		/// </summary>
		//
	    //******************************************************************
		public RepPP.Group TemplateGroup {
		    get {
		        return(m_grpTemplate);
		    }
		    set {
		        if (m_grpTemplate != value) {
		            if (m_ginstParam != null) {
		                m_ginstParam.Delete();
		                m_ginstParam = null;
		            }
		            if (m_grpParam != null) {
		                m_grpParam.Dispose();
		                m_grpParam = null;
		            }
		            if (value != null) {
		                m_grpParam    = BuildNewGroup("$HIDE_FILT" + (m_iGlobCounter++).ToString(), value);
		                m_ginstParam  = m_grpParam.GroupInstances.Create(RepPP.GroupType.sdVertical);
		            } else {
		                m_grpParam    = null;
		                m_ginstParam  = null;
		            }
		            m_grpTemplate = value;
		        }
		    }
		}

	    //******************************************************************
	    //
		/// <summary>
		/// Between:            String used to express the BETWEEN
		/// </summary>
		//
	    //******************************************************************
		public string Between {
		    get {
		        return(m_strBetween);
		    }
		    set {
                m_strBetween = value;
            }
        }

	    //******************************************************************
	    //
		/// <summary>
		/// WildCharAny:           Wild any char for LIKE clause
		/// </summary>
		//
	    //******************************************************************
		public char WildCharAny {
		    get {
		        return(m_chrWildCharAny);
		    }
		    set {
                m_chrWildCharAny = value;
            }
        }

	    //******************************************************************
	    //
		/// <summary>
		/// WildCharOne:           Wild one char for LIKE clause
		/// </summary>
		//
	    //******************************************************************
		public char WildCharOne {
		    get {
		        return(m_chrWildCharOne);
		    }
		    set {
                m_chrWildCharOne = value;
            }
        }

	    //******************************************************************
	    //
	    /// <summary>
	    /// ParamGroupInstance:         Group instance containing the fields use as filter
	    /// </summary>
	    //
	    //******************************************************************
		public RepPP.GroupInstance ParamGroupInstance {
		    get {
		        return(m_ginstParam);
		    }
		}

	    //******************************************************************
	    //
	    /// <summary>
	    /// ShowUserFlagField:      Indicateds whether or not to show the user flag field
	    ///                         on the filter    
	    /// </summary>
	    //
	    //******************************************************************
        public bool ShowUserFlagField {
            get {
                return(m_bShowUserFlagField);
            }
            set {
                m_bShowUserFlagField = value;
            }
        }

	    //******************************************************************
	    //
		/// <summary>
		/// BuildValue:                         Build the value of a field
		/// </summary>
		/// <param name="fld">                  Field</param>
		/// <returns>
		/// Value to be used in the WHERE clause
		/// </returns>
	    //
	    //******************************************************************
		protected virtual string BuildValue(RepPP.Field fld) {
		    string  strRetVal;
		    
		    switch(fld.Type) {
		    case RepPP.FieldType.sdFieldDate:
		        strRetVal = "SD_CHAR_TO_DATE('" + fld.Value + "')";
		        break;
		    case RepPP.FieldType.sdFieldDateTime:
		        strRetVal = "SD_CHAR_TO_TIMESTAMP('" + fld.Value + "')";
		        break;
		    case RepPP.FieldType.sdFieldTime:
		        strRetVal = "SD_CHAR_TO_TIME('" + fld.Value + "')";
		        break;
		    case RepPP.FieldType.sdFieldFormatString:
		    case RepPP.FieldType.sdFieldHexa:
		    case RepPP.FieldType.sdFieldMemo:
		    case RepPP.FieldType.sdFieldMultiAlphaChoice:
            case RepPP.FieldType.sdFieldSimpleAlphaChoice:
            case RepPP.FieldType.sdFieldString:
            case RepPP.FieldType.sdFieldUpperString:
                strRetVal = "'" + fld.Value.Replace("'", "''") + "'";
                break;
		    case RepPP.FieldType.sdFieldMoney:
		    case RepPP.FieldType.sdFieldMultiNumChoice:
		    case RepPP.FieldType.sdFieldNumeric:
		    case RepPP.FieldType.sdFieldNumericDec:
		    case RepPP.FieldType.sdFieldPercentage:
		    case RepPP.FieldType.sdFieldPercentageDec:
		    case RepPP.FieldType.sdFieldSimpleNumChoice:
                strRetVal = fld.Value;
                break;
            default:
                strRetVal = null;
                break;
		    }
		    return(strRetVal);
        }

        //******************************************************************
        //
        /// <summary>
        /// BuildFieldWhereClause:          Build the WHERE clause for a field
        /// </summary>
        /// <param name="fld">              Field</param>
        /// <param name="strValue">         Value specified for the field</param>
        /// <param name="strOperator">      The operator</param>
        /// <returns>
        /// WHERE clause
        /// </returns>
        //
        //******************************************************************
        protected virtual string BuildFieldWhereClause(RepPP.Field fld, string strValue, string strOperator) {
            return (fld.ColName + strOperator + strValue);
        }

        //******************************************************************
        //
        /// <summary>
        /// BuildFieldFilterClause:         Build the filter clause for a field
        /// </summary>
        /// <param name="fld">              Field</param>
        /// <param name="strValue">         Value specified for the field</param>
        /// <param name="strOperator">      The operator</param>
        /// <returns>
        /// WHERE clause
        /// </returns>
        //
        //******************************************************************
        protected virtual string BuildFieldFilterClause(RepPP.Field fld, string strValue, string strOperator) {
            return (fld.Name + strOperator + "'" + strValue + "'");
        }

	    //******************************************************************
	    //
		/// <summary>
		/// BuildWhereClause:                   Build a WHERE clause using the values in the group
		/// </summary>
		/// <param name="eType">                AND or OR</param>
		/// <param name="arrOp">                Array of operators</param>
		/// <param name="strFilterClause">      Human readable WHERE clause</param>
		/// <returns>
		/// WHERE clause
		/// </returns>
	    //
	    //******************************************************************
		protected virtual string BuildWhereClause(AndOrE eType, string[] arrOp, out string strFilterClause) {
		    string  strWhereClause;
		    string  strVal;
		    string  strVal2;
		    string  strOP;
		    int     iPos = 0;
		    
		    strWhereClause  = String.Empty;
		    strFilterClause = String.Empty;
		    foreach (Field fld in m_arrFld) {
		        if (!fld.IsNull && (!fld.IsChoiceAvailable || fld.Value != m_strNoSelection)) {
		            strVal = BuildValue(fld);
		            if (strVal != null) {
		                if (strWhereClause != String.Empty) {
		                    strWhereClause  += (eType == AndOrE.AND) ? " AND " : " OR ";
		                    strFilterClause += (eType == AndOrE.AND) ? " AND " : " OR ";
		                }
		                strOP = arrOp[iPos];
		                if (strOP == m_strBetween) {
		                    strVal2          = BuildValue(m_arrBetweenFld[iPos]);
                            strWhereClause  += BuildFieldWhereClause(fld, strVal + " AND " + strVal2, " BETWEEN ");
                            strFilterClause += BuildFieldFilterClause(fld, fld.ExtValue + "' AND '" + m_arrBetweenFld[iPos].ExtValue + "'", " BETWEEN ");
		                } else {
		                    if (strOP == "=" || strOP == "<>") {
		                        switch(fld.Type) {
		                        case RepPP.FieldType.sdFieldString:
		                        case RepPP.FieldType.sdFieldUpperString:
		                            if (strVal.IndexOf(m_chrWildCharAny) != -1 ||
		                                strVal.IndexOf(m_chrWildCharAny) != -1) {
		                                strOP  = (strOP == "=") ? " LIKE " : " NOT LIKE ";
                                        if (m_chrWildCharAny != '%') {
                                            strVal = strVal.Replace(m_chrWildCharAny.ToString(), "%");
                                        }
                                        if (m_chrWildCharOne != '_') {
                                            strVal = strVal.Replace(m_chrWildCharOne.ToString(), "_");
                                        }
		                            }
		                            break;
		                        default:
		                            break;
		                        }
		                    }
                            strWhereClause   += BuildFieldWhereClause(fld, strVal, strOP);
                            strFilterClause  += BuildFieldFilterClause(fld, fld.ExtValue, strOP);
  		                }
  		            }
  		        }
  		        iPos++;
		    }
		    if (strWhereClause != String.Empty) {
		        strWhereClause = "(" + strWhereClause + ")";
		    }
		    return(strWhereClause);
		}

	    //******************************************************************
	    //
		/// <summary>
		/// BuildWhereClause:                   Build a WHERE clause using the values in the group
		/// </summary>
		/// <param name="eType">                AND or OR</param>
		/// <param name="strFilterClause">      Human readable WHERE clause</param>
		/// <returns>
		/// WHERE clause
		/// </returns>
	    //
	    //******************************************************************
	    protected abstract string BuildWhereClause(AndOrE eType, out string strFilterClause);

	    //******************************************************************
	    //
		/// <summary>
		/// WhereClause:                    Actual WHERE clause
		/// </summary>
		/// <returns>
		/// WHERE clause
		/// </returns>
	    //
	    //******************************************************************
		public virtual string WhereClause {
		    get {
		        return(m_strWhereClause);
		    }
		    set {
		        m_strWhereClause = value;
		    }
		}

	    //******************************************************************
	    //
		/// <summary>
		/// FilterClause:                    Human readable WHERE clause
		/// </summary>
		/// <returns>
		/// Filter clause
		/// </returns>
	    //
	    //******************************************************************
		public virtual string FilterClause {
		    get {
		        return(m_strFilterClause);
		    }
		    set {
		        m_strFilterClause = value;
		    }
		}

	    //******************************************************************
	    //
		/// <summary>
		/// SetFilter:                      Set the filter
		/// </summary>
		/// <param name="eType">            Use AND or OR</param>
		//
	    //******************************************************************
		public virtual void SetFilter(AndOrE eType) {
		    m_strWhereClause = BuildWhereClause(eType, out m_strFilterClause);
		}

	    //******************************************************************
	    //
		/// <summary>
		/// AddFilter:                      Add a filter
		/// </summary>
		/// <param name="eType">            Use AND or OR</param>
		//
	    //******************************************************************
		public virtual void AddFilter(AndOrE eType) {
		    string  strWhereClause;
		    
		    strWhereClause = BuildWhereClause(eType, out m_strFilterClause);
		    if (m_strWhereClause == null || m_strWhereClause == String.Empty) {
		        m_strWhereClause = strWhereClause;
		    } else {
		        m_strWhereClause += " AND " + strWhereClause;
		    }
		}

        //******************************************************************
        //
        /// <summary>
        /// ReadCardUsingFilterV        Read the card using the filter
        /// </summary>
        /// <param name="cinst">        Card instance</param>
        /// <param name="strAddWhere">  Additionnal where clause used to filter the card</param>
        //
        //******************************************************************
        public virtual void ReadCardUsingFilterV(RepPP.CardInstance cinst, string strAddWhere) {
            ReadCardUsingFilterV(cinst, strAddWhere, null);
        }

        //******************************************************************
        //
        /// <summary>
        /// ReadCardUsingFilterV        Read the card using the filter
        /// </summary>
        /// <param name="cinst">        Card instance</param>
        /// <param name="strAddWhere">  Additionnal where clause used to filter the card</param>
        /// <param name="dataAccess">   Data access object</param>
        //
        //******************************************************************
        public virtual void ReadCardUsingFilterV(RepPP.CardInstance     cinst, 
                                                 string                 strAddWhere,
                                                 ISBDataAccess          dataAccess) {
            string strWhereClause = String.Empty;

            if (strAddWhere != null && strAddWhere != String.Empty) {
                strWhereClause = strAddWhere;
            }
            if (WhereClause != String.Empty) {
                strWhereClause += (strWhereClause == String.Empty) ? "WHERE " + WhereClause :
                                                                    " AND (" + WhereClause + ")";
            }
            cinst.Card.ChildCardGroups[0].SetSelectSqlWhere(strWhereClause);
            if (dataAccess != null) {
                dataAccess.LoadSelectionBufferData(cinst, strWhereClause);
            } else {
                cinst.ReadFromDb();
            }
        }

	    //******************************************************************
	    //
	    /// <summary>
	    /// IsValid:            Checks if the value are valids
	    /// </summary>
	    //
	    //******************************************************************
		public bool IsValid() {
		    bool    bRetVal = true;
		    		    
		    foreach (RepPP.Field fld in m_grpParam.Fields) {
		        if (fld.Validate() != 0) {
		            bRetVal = false;
		        }
		    }
		    return(bRetVal);
        }
    }
}
