//*********************************************************
//
// DynamicForm          Handle dynamic form creation
//
// Author:		        Jacques Fournier
//
// (c) 2003 by Consyst SQL Inc.
//
//
//*********************************************************
using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Windows.Forms;
using RepPP.Toolkit;

namespace RepPP.Toolkit.Window {
    /// <summary>
    /// Handle dynamic form creation
    /// </summary>
    public class DynamicForm {
        /// <summary>Options for form generation</summary>
        public enum OptionE {
            /// <summary>No option</summary>
            OPT_NONE                            = 0,
            /// <summary>Only add field attached to the main table</summary>
            OPT_ONLY_MAIN_TBL_FIELD             = 1,
            /// <summary>Field are always selectable</summary>
            OPT_ALWAYS_SELECTABLE               = 2,
            /// <summary>Replace radio by a combo box</summary>
            OPT_NO_RADIO                        = 4,
            /// <summary>Replace check by listbox</summary>
            OPT_NO_CHECK                        = 8,
            /// <summary>Add a combo box with operators</summary>
            OPT_ADD_OPERATOR_COMBO              = 16,
            /// <summary>Add an additional invisible field for BETWEEN clause</summary>
            OPT_ADD_BETWEEN_FIELD               = 32,
            /// <summary>Don't add fields which are user flsg</summary>
            OPT_NO_USER_FLAG_FIELD              = 64
        };
        /// <summary>
        /// Default control height
        /// </summary>
        private const int   DEF_CTL_HEIGHT      =   14;
        /// <summary>
        /// Horizontal space between controls
        /// </summary>
        private const int   HORZ_MARGE_SPACE    =   4;
        /// <summary>
        /// Vertical space between controls
        /// </summary>
        private const int   VERT_MARGE_SPACE    =   4;
        /// <summary>
        /// Size of the bitmap for the combobox, radio or check
        /// </summary>
        private const int   BUTTON_BITMAP_SIZE  =   12;

        //*********************************************************	    
        //
        /// <summary>
        /// ToPixelX:               Convert our logical unit in pixel
        /// </summary>
        /// <param name="gr">       Graphic object</param>
        /// <param name="x">        X position to translate</param>
        /// <param name="font">     font used</param>
        /// <returns>
        /// Position in pixels
        /// </returns>
        //
        //*********************************************************	    
        static private int ToPixelX(System.Drawing.Graphics gr, int x, System.Drawing.Font font)
        {
            return((int)(gr.MeasureString("H", font).Width * x / 4));
        }

        //*********************************************************	    
        //
        /// <summary>
        /// ToPixelY:               Convert our logical unit in pixel
        /// </summary>
        /// <param name="gr">       Graphic object</param>
        /// <param name="y">        Y position to translate</param>
        /// <param name="font">     font used</param>
        /// <returns>
        /// Position in pixels
        /// </returns>
        //
        //*********************************************************	    
        static private int ToPixelY(System.Drawing.Graphics gr, int y, System.Drawing.Font font)
        {
            return((int)(font.GetHeight(gr) * y / 8));
        }

        //*********************************************************	    
        //
        /// <summary>
        /// ComputeMaxPromptSizei:  Compute the maximum prompt size
        /// </summary>
        /// <param name="gr">       Graphic object</param>
        /// <param name="grp">      Group to use as template</param>
        /// <param name="font">     Font to use</param>
        //
        //*********************************************************	    
        static private int ComputeMaxPromptSizei(System.Drawing.Graphics gr, RepPP.Group grp, System.Drawing.Font font)
        {
            int                     iRetVal = 0;
            int                     iPromptSize;
            
            foreach (RepPP.Field fld in grp.Fields) {
                iPromptSize = (int)gr.MeasureString(fld.Prompt, font).Width;
                if (iPromptSize > iRetVal) {
                    iRetVal = iPromptSize;
                }
            }
            return(iRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// ComputeMaxChoiceSizei:      Compute the maximum choice size
        /// </summary>
        /// <param name="gr">           Graphic object</param>
        /// <param name="choiceList">   Choice list</param>
        /// <param name="font">         Font to use</param>
        /// <param name="bCodeToo">     true to add the code and tab between code and description</param>
        //
        //*********************************************************	    
        static private int ComputeMaxChoiceSizei(System.Drawing.Graphics gr, RepPP.ChoiceList choiceList, System.Drawing.Font font, bool bCodeToo)
        {
            int                     iRetVal = 0;
            int                     iChoiceSize;
            int                     iChoiceCount;
            string                  strCode;
            string                  strDesc;
            string                  str;
            
            iChoiceCount = choiceList.Count;
            for (int i = 0; i < iChoiceCount; i++) {
                strCode = choiceList.GetCode(i);
                strDesc = choiceList.GetDesc(i);
                if (bCodeToo) {
                    if (strDesc == null || strDesc.Length == 0) {
                        str = strCode;
                    } else {
                        str = strCode + "\t" + strDesc;
                    }
                } else {
                    str = (strDesc == null || strDesc.Length == 0) ? strCode : strDesc;
                }
                iChoiceSize = (int)gr.MeasureString(str, font).Width;
                if (iChoiceSize > iRetVal) {
                    iRetVal = iChoiceSize;
                }
            }
            return(iRetVal);
        }

        //*********************************************************	    
        //
        /// <summary>
        /// CreateGroupFields:                  Create field controls in the specified container.
        /// </summary>
        /// <param name="ctl">                  Control in which to add the controls</param>
        /// <param name="grp">                  Group to use as template</param>
        /// <param name="font">                 Font to use</param>
        /// <param name="iMaxPromptSizeInChr">  Maximum prompt size in characters</param>
        /// <param name="iMaxFieldSizeInChr">   Maximum field size in characters</param>
        /// <param name="iComboBoxWidthInChr">  Width of combobox if option is selected</param>
        /// <param name="iSpaceAfterComboInChr">Space after combo in characters</param>
        /// <param name="eOption">              Generation option</param>
        /// <param name="iMaxXPos">             Returned the maximum right edge of a field</param>
        /// <param name="arrComboBox">          Array of combo box generated if option specified</param>
        /// <param name="arrStat">              Array of created static fields</param>
        /// <param name="arrField">             Array of created fields</param>
        /// <param name="arrBetweenField">      Array of invisible fields created for the BEWTEEN clause</param>
        //
        //*********************************************************	    
        static public void CreateGroupFields(System.Windows.Forms.Control   ctl,
                                             RepPP.Group                    grp,
                                             System.Drawing.Font            font, 
                                             int                            iMaxPromptSizeInChr,
                                             int                            iMaxFieldSizeInChr,
                                             int                            iComboBoxWidthInChr,
                                             int                            iSpaceAfterComboInChr,
                                             OptionE                        eOption,
                                             out int                        iMaxXPos,
                                             out ComboBox[]                 arrComboBox,
                                             out RepPP.RepPPStat[]          arrStat,
                                             out RepPP.RepPPField[]         arrField,
                                             out RepPP.RepPPField[]         arrBetweenField)
        {
            int                 iMaxPromptSize;
            Graphics            gr;
            RepPP.CtlType       eCtlType;
            RepPP.RepPPStat     statCtl;
            RepPP.RepPPField    fldCtl;
            int                 iYPos;
            int                 iYOldPos;
            int                 iStaticYSize;
            int                 iStaticXPos;
            int                 iFieldXPos;
            int                 iFieldAdjustXPos;
            int                 iFieldXSize;
            int                 iFieldYSize;
            int                 iMaxFieldXSize;
            int                 iMaxFieldXSizeDiv2;
            int                 iXMargin;
            int                 iYMargin;
            int                 iButtonXSize;   // Use for combo, radio or check
            int                 iComboBoxWidth;
            int                 iSpaceAfterCombo;
            int                 iTmpSize;
            ArrayList           listCombo;
            ArrayList           listStat;
            ArrayList           listField;
            ArrayList           listBetweenField;
            ComboBox            combo;
            bool                bBetweenField;
            
            if (font == null) {
                font = ctl.Font;
            }
            listCombo       = new ArrayList(grp.Fields.Count);
            listStat        = new ArrayList(grp.Fields.Count);
            listField       = new ArrayList(grp.Fields.Count);
            listBetweenField= new ArrayList(grp.Fields.Count);
            gr              = ctl.CreateGraphics();
            iXMargin        = ToPixelX(gr, HORZ_MARGE_SPACE, font);
            iYMargin        = ToPixelY(gr, VERT_MARGE_SPACE, font);
            iTmpSize        = (int)gr.MeasureString("M".PadRight(iMaxPromptSizeInChr, 'M'), font).Width;
            iMaxPromptSize  = ComputeMaxPromptSizei(gr, grp, font);
            iComboBoxWidth  = (int)gr.MeasureString("M".PadRight(iComboBoxWidthInChr, 'M'), font).Width;
            iSpaceAfterCombo= (int)gr.MeasureString("M".PadRight(iSpaceAfterComboInChr, 'M'), font).Width;
            if (iMaxPromptSize > iTmpSize) {
                iMaxPromptSize = iTmpSize;
            }
            iMaxFieldXSize      = (int)gr.MeasureString("M".PadRight(iMaxFieldSizeInChr, 'M'), font).Width;
            iMaxFieldXSizeDiv2  = iMaxFieldXSize / 2 - 1;
            iStaticYSize        = ToPixelY(gr, DEF_CTL_HEIGHT, font);
            iStaticXPos         = iXMargin;
            iFieldXPos          = iStaticXPos + iMaxPromptSize + iXMargin;
            iYPos               = iYMargin;
            iButtonXSize        = ToPixelX(gr, BUTTON_BITMAP_SIZE, font);
            iMaxXPos            = iFieldXPos + iComboBoxWidth + iSpaceAfterCombo + iMaxFieldXSize;
            iYOldPos            = iYPos;
            foreach (RepPP.Field fld in grp.Fields) {
                bBetweenField = (fld.Name[0] == '_');
                if (fld.IsVisible || 
                    ((eOption & OptionE.OPT_ADD_BETWEEN_FIELD) != 0 && bBetweenField)) {
                    if (fld.IsDatabaseLinked ||
                        (eOption & OptionE.OPT_ONLY_MAIN_TBL_FIELD) == 0) {
                        if (!fld.IsUserFlagOn ||
                            (eOption & OptionE.OPT_NO_USER_FLAG_FIELD) == 0) {
                            eCtlType    = fld.ControlType;
                            switch(eCtlType) {
                            case RepPP.CtlType.sdCtlCheck:
                                if ((eOption & OptionE.OPT_NO_CHECK) != 0) {
                                    eCtlType = RepPP.CtlType.sdCtlCombo;
                                }
                                break;
                            case RepPP.CtlType.sdCtlRadio:
                                if ((eOption & OptionE.OPT_NO_RADIO) != 0) {
                                    eCtlType = RepPP.CtlType.sdCtlList;
                                }
                                break;
                            case RepPP.CtlType.sdCtlEdit:
                                if (fld.ChoiceList.Count > 0) {
                                    eCtlType = RepPP.CtlType.sdCtlCombo;
                                }
                                break;
                            }
                            iFieldXSize = 0;
                            switch(eCtlType) {
                            case RepPP.CtlType.sdCtlCheck:
                            case RepPP.CtlType.sdCtlRadio:
                                iFieldXSize = ComputeMaxChoiceSizei(gr, fld.ChoiceList, font, false) + iButtonXSize;
                                iFieldYSize = (iStaticYSize + iYMargin) * fld.ChoiceList.Count - iYMargin;
                                break;
                            case RepPP.CtlType.sdCtlCombo:
                                iFieldXSize = ComputeMaxChoiceSizei(gr, fld.ChoiceList, font, true) + iButtonXSize;
                                iFieldYSize = iStaticYSize;
                                break;
                            case RepPP.CtlType.sdCtlList:
                                iFieldXSize = ComputeMaxChoiceSizei(gr, fld.ChoiceList, font, true) + iButtonXSize;
                                iFieldYSize = (iStaticYSize + iYMargin) * 4 - iYMargin;
                                break;
                            case RepPP.CtlType.sdCtlFmt:
                            case RepPP.CtlType.sdCtlFmtDate:
                            case RepPP.CtlType.sdCtlMChr:
                            case RepPP.CtlType.sdCtlMLine:
                                iFieldYSize = iStaticYSize;
                                break;
                            case RepPP.CtlType.sdCtlEdit:
                            default:
                                eCtlType    = RepPP.CtlType.sdCtlEdit;
                                iFieldYSize = (fld.Type == RepPP.FieldType.sdFieldMemo) ? (iStaticYSize + iYMargin) * 4 - iYMargin : iStaticYSize;
                                break;
                            }
                            if (iFieldXSize == 0) {
                                iTmpSize    = (fld.MaxSize < iMaxFieldSizeInChr) ? fld.MaxSize : iMaxFieldSizeInChr;
                                iFieldXSize = (int)gr.MeasureString("M".PadRight(iTmpSize, 'M'), font).Width;
                            }
                            if (bBetweenField) {
                                if (iFieldXSize > iMaxFieldXSizeDiv2) {
                                    iFieldXSize = iMaxFieldXSizeDiv2;
                                }
			                    iFieldAdjustXPos  = iFieldXPos + 2;
			                    if ((eOption & OptionE.OPT_ADD_OPERATOR_COMBO) != 0) {
			                        iFieldAdjustXPos += iComboBoxWidth + iSpaceAfterCombo + iFieldXSize;
                                }
                            } else {
                                if (iFieldXSize > iMaxFieldXSize) {
                                    iFieldXSize = iMaxFieldXSize;
                                }
			                    statCtl                 = new RepPP.RepPPStat();
			                    statCtl.AtomName        = "";
			                    statCtl.AtomType        = RepPP.AtomType.sdAtomField;
			                    statCtl.BorderStyle     = System.Windows.Forms.BorderStyle.None;
			                    statCtl.ControlType     = RepPP.CtlType.sdCtlStatic;
			                    statCtl.FieldName       = fld.Name;
			                    statCtl.Font            = font;
			                    statCtl.ForeColor       = System.Drawing.Color.Black;
			                    statCtl.GroupName       = fld.Group.Name;
			                    statCtl.Location        = new System.Drawing.Point(iStaticXPos, iYPos);
			                    statCtl.Name            = "sta" + fld.Name;
			                    statCtl.Prefix          = false;
			                    statCtl.Size            = new System.Drawing.Size(iMaxPromptSize, iStaticYSize);
			                    statCtl.TabIndex        = 0;
			                    statCtl.TabStop         = false;
			                    statCtl.TextAlign       = System.Windows.Forms.HorizontalAlignment.Left;
			                    ctl.Controls.Add(statCtl);
			                    listStat.Add(statCtl);
			                    if ((eOption & OptionE.OPT_ADD_OPERATOR_COMBO) != 0) {
			                        combo            = new ComboBox();
			                        combo.Name       = "combo" + fld.Name;
			                        combo.Location   = new System.Drawing.Point(iFieldXPos, iYPos);
			                        combo.Size       = new System.Drawing.Size(iComboBoxWidth, iFieldYSize);
			                        combo.TabIndex   = fld.LogicalPos;
			                        combo.TabStop    = true;
			                        iFieldAdjustXPos = iFieldXPos + iComboBoxWidth + iSpaceAfterCombo;
			                        listCombo.Add(combo);
			                        ctl.Controls.Add(combo);
			                    } else {
			                        iFieldAdjustXPos = iFieldXPos;
			                    }
			                }
 			                fldCtl                  = new RepPP.RepPPField();
			                fldCtl.AutoHScroll      = true;
			                fldCtl.AutoVScroll      = false;
			                fldCtl.BackColor        = System.Drawing.SystemColors.Window;
			                fldCtl.BorderStyle      = System.Windows.Forms.BorderStyle.Fixed3D;
			                fldCtl.ComboBoxStyle    = System.Windows.Forms.ComboBoxStyle.DropDownList;
			                fldCtl.ControlType      = eCtlType;
			                fldCtl.FieldName        = fld.Name;
			                fldCtl.Font             = font;
			                fldCtl.ForeColor        = System.Drawing.SystemColors.Desktop;
			                fldCtl.GroupName        = fld.Group.Name;
			                fldCtl.HideSelection    = true;
			                fldCtl.LeftText         = false;
			                fldCtl.Location         = new System.Drawing.Point(iFieldAdjustXPos, (bBetweenField) ? iYOldPos : iYPos);
			                fldCtl.Multiline        = (fld.Type == RepPP.FieldType.sdFieldMemo) ? true : false;
			                fldCtl.Name             = "fld" + fld.Name;
			                fldCtl.Placement        = RepPP.Placement.sdLeftToRightWrap;
			                fldCtl.ScrollBars       = System.Windows.Forms.ScrollBars.None;
			                fldCtl.Size             = new System.Drawing.Size(iFieldXSize, iFieldYSize);
			                fldCtl.TabIndex         = fld.LogicalPos + 1;
			                fldCtl.TabStop          = true;
			                fldCtl.TextAlign        = System.Windows.Forms.HorizontalAlignment.Left;
			                fldCtl.Visible          = fld.IsVisible;
                            if (eCtlType == CtlType.sdCtlCombo && fld.ControlType == CtlType.sdCtlEdit) {
                                fldCtl.ComboBoxStyle = ComboBoxStyle.DropDown;
                            }
			                ctl.Controls.Add(fldCtl);
			                iYOldPos  = iYPos;
			                if (bBetweenField) {
			                    listBetweenField.Add(fldCtl);
			                } else {
			                    listField.Add(fldCtl);
			                    iYPos    += iFieldYSize + iYMargin;
			                }
                        }
			        }
                }
            }
            arrComboBox         = (ComboBox[])listCombo.ToArray(typeof(ComboBox));
            arrStat             = (RepPP.RepPPStat[])listStat.ToArray(typeof(RepPP.RepPPStat));
            arrField            = (RepPP.RepPPField[])listField.ToArray(typeof(RepPP.RepPPField));
            arrBetweenField     = (RepPP.RepPPField[])listBetweenField.ToArray(typeof(RepPP.RepPPField));
        }

        //*********************************************************	    
        //
        /// <summary>
        /// CreateGroupFields:                  Create field controls in the specified container.
        /// </summary>
        /// <param name="ctl">                  Control in which to add the controls</param>
        /// <param name="grp">                  Group to use as template</param>
        /// <param name="font">                 Font to use</param>
        /// <param name="iMaxPromptSizeInChr">  Maximum prompt size in characters</param>
        /// <param name="iMaxFieldSizeInChr">   Maximum field size in characters</param>
        //
        //*********************************************************	    
        static public void CreateGroupFields(System.Windows.Forms.Control ctl, RepPP.Group grp, System.Drawing.Font font, int iMaxPromptSizeInChr, int iMaxFieldSizeInChr)
        {
            ComboBox[]          arrDummy;
            RepPP.RepPPStat[]   arrDummy2;
            RepPP.RepPPField[]  arrDummy3;
            RepPP.RepPPField[]  arrDummy4;
            int                 iDummy;
            
            CreateGroupFields(ctl, grp, font, iMaxPromptSizeInChr, iMaxFieldSizeInChr, 0, 0, OptionE.OPT_NONE, out iDummy, out arrDummy, out arrDummy2, out arrDummy3, out arrDummy4);
        }
    }
}