using System;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using Win32;
namespace RepPP.Toolkit.Window {

    //*********************************************************	    
    //
    /// <summary>
    /// Implement a tree view attached containing TreeNodeBase or derived node.
    /// </summary>
    //
    //*********************************************************	
    [ToolboxBitmap(typeof(VComboBox),"FldComboBox.bmp")]
    [DefaultEvent("SelectedIndexChanged")]
    public class VComboBox : Control {
        /// <summary>Argument for GetItemText and FindItemText event</summary>
        public class ItemTextArg : System.EventArgs {
            /// <summary>Index of the item</summary>
            public int                          Index;
            /// <summary>Text of the item</summary>
            public string                       Text;
            /// <summary>
            ///     ctor for GetItemText
            /// </summary>
            /// <param name="iIndex">Index of the item</param>
            public                              ItemTextArg(int iIndex) { Index = iIndex; Text = iIndex.ToString(); }
            /// <summary>
            /// ctor for FindItem
            /// </summary>
            /// <param name="strText">Text to find</param>
            public                              ItemTextArg(string strText) { Text = strText; Index = -1; }
        }

        /// <summary>
        /// Delegate for GetItemText and FindItemText event
        /// </summary>
        /// <param name="sender">   Sender object</param>
        /// <param name="e">        Event argument</param>
        public delegate void                    ItemTextEventHandler(object sender, ItemTextArg e);
        /// <summary>Event sent to get the text at the specified item index</summary>
        public event ItemTextEventHandler       GetItemText;
        /// <summary>Find the closest matching string in the list. Use for AutoComplete</summary>
        public event ItemTextEventHandler       FindItemIndex;

        [StructLayout(LayoutKind.Sequential, CharSet=CharSet.Auto)]
        private class VIRTUALCOMBOBOXITEM {
            public IntPtr   hwndFrom = IntPtr.Zero;
            public IntPtr   idFrom = IntPtr.Zero;
            public int      code = 0;
            public int      iIndex = 0;
            public IntPtr   lpszText = IntPtr.Zero;
        }

        private enum VCBN : int {
            GETDISPINFO     =   2,
            DISPCHANGE      =   3
        }

        private enum VCBCMDN : int {
            CLOSEUP			=   1,
            DBLCLK		    =	2,
            DROPDOWN        =	3,
            EDITCHANGE		=	4,
            EDITUPDATE		=	5,
            KILLFOCUS		=	6,
            SELCHANGE		=	7,
            SETFOCUS		=	8,
        }

        private enum VCBM : uint {            
            FIRST			        = 0x0401,
            GETCOUNT				= 0x0401,
            GETCURSEL				= 0x0402,
            GETDROPPEDCONTROLRECT	= 0x0403,
            GETDROPPEDSTATE		    = 0x0404,
            GETEDITSEL				= 0x0405,
            GETTOPINDEX			    = 0x0406,
            LIMITTEXT				= 0x0407,
            SETCOUNT				= 0x0408,
            SETCURSEL				= 0x0409,
            SETEDITSEL				= 0x040A,
            SETTOPINDEX			    = 0x040B,
            SHOWDROPDOWN			= 0x040C,
            SETDROPPEDWIDTH         = 0x040D,   // TODO
            SETEDITSELTEXT          = 0x40E     // TODO
        };
        
        /// <summary>Occurs when the drop-down portion of a ComboBox is shown.</summary>
        public event    EventHandler    DropDown;
        /// <summary>Occurs when the DropDownStyle property has changed.</summary>
        public event    EventHandler    DropDownStyleChanged;
        /// <summary>Occurs when the SelectedIndex property has changed.</summary>
        public event    EventHandler    SelectedIndexChanged;
        /// <summary>Occurs when the selected item changed</summary>
        public event    EventHandler    SelectedItemChanged;
        /// <summary>Occurs when the SelectedValue property changes.</summary>
        public event    EventHandler    SelectedValueChanged;
        /// <summary>Occurs when the selected item has changed and that change is displayed in the ComboBox.</summary>
        public event    EventHandler    SelectionChangeCommitted;
        
        private GCHandle                m_childWindowRoot;          // GCHandle to the child window proc
        private int                     m_iSelectedIndex;           // Selected index (before handle created)
        private ComboBoxStyle           m_eComboStyle;              // Combo box style
        private int                     m_iDropDownWidth;           // Width of the listbox when opened
        private int                     m_iMaxDropDownItems;        // Maximum nb of shown drop down items
        private int                     m_iMaxLength;               // Maximum text accepted in the edit text
        private int                     m_iPreferredHeight;         // Preferred height when the combo box is created
        private IntPtr                  m_wndListBox;               // Listbox window handle
        private IntPtr                  m_wndEdit;                  // Editbox window handle
        private IntPtr                  m_wndDropDown;              // Drop down window handle (Always the same as Listbox?)
        private IntPtr                  m_procListBox;              // Original window proc of the listbox
        private IntPtr                  m_procEdit;                 // Original window proc of the edit text
        private bool                    m_bCanFireLostFocus;        // true if can fire the lost focus event
        private bool                    m_bFireLostFocus;           // Firing the lost focus event
        private bool                    m_bFireSetFocus;            // Firing the set focus event
        private bool                    m_bIntegralHeight;          // Indicates whether the control should resize to avoid showing partial items
        private int                     m_iRequestedHeight;         // Requested height of the combo box
        private int                     m_iItemCount;               // Nb of items in the combo box
        private bool                    m_bMousePressed = false;    // If a mouse button is pressed
        private bool                    m_bMouseEvents = false;     // If in a mouse event
        private bool                    m_bMouseInEdit = false;     // If mouse in edit text
        private IntPtr                  m_brushBack;                // Background brush
        private bool                    m_bAutoComplete;            // true if autocomplete is on
        private bool                    m_bAutoDropDown;            // true if autodropdown is on
        private bool                    m_bProcessAutoComplete;     // true if next text change must trigger auto-complete

        //********************************************************************************
        //
        /// <summary>
        /// ctor:                       Class constructor
        /// </summary>
        //
        //********************************************************************************
        public VComboBox() {
            m_iSelectedIndex                = -1;
            m_iPreferredHeight              = -1;
            m_bIntegralHeight               = true;
            m_bFireSetFocus                 = true;
            m_bFireLostFocus                = true;
            m_bCanFireLostFocus             = false;
            m_iRequestedHeight              = 150;
            m_eComboStyle                   = ComboBoxStyle.DropDown;
            m_iDropDownWidth                = -1;
            m_iMaxDropDownItems             = 8;
            m_iMaxLength                    = 0;
            m_wndListBox                    = IntPtr.Zero;
            m_wndEdit                       = IntPtr.Zero;
            m_procListBox                   = IntPtr.Zero;
            m_procEdit                      = IntPtr.Zero;
            m_wndDropDown                   = IntPtr.Zero;
            m_bAutoComplete                 = false;
            m_bAutoDropDown                 = false;
            m_bProcessAutoComplete          = false;
            BackColor                       = SystemColors.Window;
            ForeColor                       = SystemColors.WindowText;
            m_iItemCount                    = 0;
            m_brushBack                     = (IntPtr)0;
            SetStyle(ControlStyles.CacheText, false);
            SetStyle(ControlStyles.StandardClick | ControlStyles.UserPaint, false);
        }

        //********************************************************************************
        //
        /// <summary>
        /// SendMessage:                Send a message to a window
        /// </summary>
		/// <param name="eMsg">          Message to send</param>
		/// <param name="wParam">       First parameter</param>
		/// <param name="lParam">       Second parameter</param>
		/// <returns>
        /// Result
        /// </returns>
        //
        //********************************************************************************
        private IntPtr SendMessage(VCBM eMsg, int wParam, int lParam) {
            return((IntPtr)User32.SendMessage(Handle, (int)eMsg, wParam, lParam));
        }

        //********************************************************************************
        //
        /// <summary>
        /// SendMessage:                Send a message to a window
        /// </summary>
		/// <param name="eMsg">          Message to send</param>
		/// <param name="iVal1">        First value</param>
		/// <param name="iVal2">        Second value</param>
		/// <returns>
        /// Result
        /// </returns>
        //
        //********************************************************************************
        private IntPtr SendMessage(VCBM eMsg, ref int iVal1, ref int iVal2) {
            return(User32.SendMessage(Handle, (int)eMsg, ref iVal1, ref iVal2));
        }

        //********************************************************************************
        //
        /// <summary>
        /// SendMessage:                Send a message to a window
        /// </summary>
		/// <param name="eMsg">         Message to send</param>
		/// <param name="wParam">       First parameter</param>
		/// <param name="strParam">     Second parameter</param>
		/// <returns>
        /// Result
        /// </returns>
        //
        //********************************************************************************
        private IntPtr SendMessage(VCBM eMsg, int wParam, string strParam) {
            return(User32.SendMessageW(Handle, (int)eMsg, wParam, strParam));
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnFindItemIndex:            Find the index of an item using a partial text
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnFindItemIndex(ItemTextArg e) {
            if (FindItemIndex != null) {
                FindItemIndex(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnGetItemText:              Gets the text of an item
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnGetItemText(ItemTextArg e) {
            if (GetItemText != null) {
                GetItemText(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// GetItemTextFromIndex:       Gets the text of an item
        /// </summary>
        /// <param name="iIndex">       Index of the item</param>
        //
        //********************************************************************************
        private string GetItemTextFromIndex(int iIndex) {
            ItemTextArg e;

            e = new ItemTextArg(iIndex);
            OnGetItemText(e);
            return(e.Text);
        }

        //********************************************************************************
        //
        /// <summary>
        /// ChildWndProc:               Child window procedure
        /// </summary>
        /// <param name="m">            Message structure</param>
        //
        //********************************************************************************
        private void ChildWndProc(ref Message m) {
            int                 iMsgId = m.Msg;
            ContainerControl    ctlContainer;
            Point               pt;

            switch((Win32.WM)iMsgId) {
            case WM.SETFOCUS:
                //if (!DesignMode) {
                //    Control.SetImeModeToIMEContext(base.CachedImeMode, m.HWnd);
                //}
                ctlContainer = GetContainerControl() as ContainerControl;
                if (ctlContainer != null) {
                    ctlContainer.ActiveControl = this;
                    if (ctlContainer.ActiveControl != this) {
                        return;
                    }
                }
                DefChildWndProc(ref m);
                if (m_bFireSetFocus) {
                    OnGotFocus(EventArgs.Empty);
                }
                break;
            case WM.KILLFOCUS:
                //if (!DesignMode) {
                //    UpdateCachedImeMode(m.HWnd);
                //}
                WndProc(ref m);
                DefChildWndProc(ref m);
                if (m_bFireLostFocus) {
                    OnLostFocus(EventArgs.Empty);
                }
                break;
            case WM.SETFONT:
                DefChildWndProc(ref m);
                if (m.HWnd == m_wndEdit) {
                    User32.SendMessage(m_wndEdit, (int)EM.SETMARGINS, 3, 0);
                }
                break;
            //case WM.GETOBJECT:
            //    ChildWmGetObject(ref m);
            //    return;
            case WM.CONTEXTMENU:
                if (ContextMenu != null) {
                    User32.SendMessage(Handle, (int)WM.CONTEXTMENU, (int)m.WParam, (int)m.LParam);
                } else {
                    DefChildWndProc(ref m);
                }
                break;
            case WM.KEYDOWN:
            case WM.SYSKEYDOWN:
            case WM.KEYUP:
            case WM.CHAR:
            case WM.SYSKEYUP:
            case WM.SYSCHAR:
                if (DropDownStyle != ComboBoxStyle.Simple || m.HWnd != m_wndListBox) {
                    if (!PreProcessMessage(ref m)) {
                        if (!ProcessKeyMessage(ref m)) {
                            DefChildWndProc(ref m);
                        }
                    }
                } else {
                    DefChildWndProc(ref m);
                }
                break;
            case WM.MOUSEMOVE:
                pt = EditToComboboxMapping(m);
                DefChildWndProc(ref m);
                OnMouseMove(new MouseEventArgs(Control.MouseButtons, 0, pt.X, pt.Y, 0));
                break;
            case WM.LBUTTONDOWN:
                m_bMousePressed = true;
                m_bMouseEvents  = true;
                Capture         = true;
                DefChildWndProc(ref m);
                OnMouseDown(new MouseEventArgs(MouseButtons.Left, 1, (short) ((int) m.LParam), ((int) m.LParam) >> 0x10, 0));
                break;
            case WM.LBUTTONUP:
                RECT    tRect;
                
                tRect = new RECT();
                User32.GetWindowRect(Handle, ref tRect);
                Rectangle rectangle1 = new Rectangle(tRect.left, tRect.top, tRect.right - tRect.left, tRect.bottom - tRect.top);
                pt = new Point((short)((int)m.LParam), ((int) m.LParam) >> 0x10);
                pt = PointToScreen(pt);
                if (m_bMouseEvents /*&& !base.ValidationCancelled */) {
                    m_bMouseEvents = false;
                    if (m_bMousePressed) {
                        m_bMousePressed = false;
                        if (!rectangle1.Contains(pt)) {
                            m_bMouseInEdit  = false;
                            OnMouseLeave(EventArgs.Empty);                            
                        } else {
                            OnClick(EventArgs.Empty);
                        }
                    }
                }
                DefChildWndProc(ref m);
                Capture = false;
                OnMouseUp(new MouseEventArgs(MouseButtons.Left, 1, (short) ((int) m.LParam), ((int) m.LParam) >> 0x10, 0));
                break;
            case WM.LBUTTONDBLCLK:
                m_bMousePressed = true;
                m_bMouseEvents  = true;
                Capture         = true;
                DefChildWndProc(ref m);
                OnMouseDown(new MouseEventArgs(MouseButtons.Left, 1, (short) ((int) m.LParam), ((int) m.LParam) >> 0x10, 0));
                break;
            case WM.RBUTTONDOWN:
                m_bMousePressed = true;
                m_bMouseEvents  = true;
                if (ContextMenu != null) {
                    Capture = true;
                }
                DefChildWndProc(ref m);
                pt = EditToComboboxMapping(m);
                OnMouseDown(new MouseEventArgs(MouseButtons.Right, 1, pt.X, pt.Y, 0));
                break;
            case WM.RBUTTONUP:
                m_bMousePressed = false;
                m_bMouseEvents  = false;
                if (ContextMenu != null) {
                    Capture = false;
                }
                DefChildWndProc(ref m);
                OnMouseUp(new MouseEventArgs(MouseButtons.Right, 1, (short) ((int) m.LParam), ((int) m.LParam) >> 0x10, 0));
                break;
            case WM.RBUTTONDBLCLK:
                m_bMousePressed = true;
                m_bMouseEvents  = true;
                Capture         = true;
                DefChildWndProc(ref m);
                OnMouseDown(new MouseEventArgs(MouseButtons.Right, 1, (short) ((int) m.LParam), ((int) m.LParam) >> 0x10, 0));
                break;
            case WM.MBUTTONDOWN:
                m_bMousePressed = true;
                m_bMouseEvents  = true;
                Capture         = true;
                DefChildWndProc(ref m);
                pt = EditToComboboxMapping(m);
                OnMouseDown(new MouseEventArgs(MouseButtons.Middle, 1, pt.X, pt.Y, 0));
                break;
            case WM.MBUTTONUP:
                m_bMousePressed = false;
                m_bMouseEvents  = false;
                Capture         = false;
                DefChildWndProc(ref m);
                OnMouseUp(new MouseEventArgs(MouseButtons.Middle, 1, (short) ((int) m.LParam), ((int) m.LParam) >> 0x10, 0));
                break;
            case WM.MBUTTONDBLCLK:
                m_bMousePressed = true;
                m_bMouseEvents  = true;
                Capture         = true;
                DefChildWndProc(ref m);
                OnMouseDown(new MouseEventArgs(MouseButtons.Middle, 1, (short) ((int) m.LParam), ((int) m.LParam) >> 0x10, 0));
                break;
            default:
                DefChildWndProc(ref m);
                break;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// DefChildWndProc:        Default child window proc
        /// </summary>
        /// <param name="m">        Message structure</param>
        //
        //********************************************************************************
        private void DefChildWndProc(ref Message m) {
            IntPtr  proc;

            proc        = (m.HWnd == m_wndEdit) ? m_procEdit : m_procListBox;
            m.Result    = User32.CallWindowProc(proc, m.HWnd, m.Msg, m.WParam, m.LParam);
        }

        //********************************************************************************
        //
        /// <summary>
        /// Dispose:                    Dispose the control
        /// </summary>
        /// <param name="bDisposing">   Disposing the control</param>
        //
        //********************************************************************************
        protected override void Dispose(bool bDisposing) {
            ReleaseChildWindow();
            base.Dispose(bDisposing);
        }

        //********************************************************************************
        //
        /// <summary>
        /// EditToComboboxMapping:      Converts a coordinate from edit to combo
        /// </summary>
        /// <param name="m">            Message structure</param>
        //
        //********************************************************************************
        private Point EditToComboboxMapping(Message m) {
            Point   ptRetVal;
            RECT    tRectCombo = new RECT();
            RECT    tRectEdit  = new RECT();
  
            ptRetVal = new Point();
            User32.GetWindowRect(Handle, ref tRectCombo);
            User32.GetWindowRect(m_wndEdit, ref tRectEdit);
            ptRetVal.X = ((short)((int) m.LParam)) + (tRectEdit.left - tRectCombo.left);
            ptRetVal.Y = (((int) m.LParam) >> 0x10) + (tRectEdit.top - tRectCombo.top);
            return(ptRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// InitializeDCForWmCtlColor:      Initialize Device context for WM_CTLCOLOR
        /// </summary>
        /// <param name="dc">               Device context</param>
        /// <param name="msg">              Message</param>
        /// <returns>
        /// Brush
        /// </returns>
        //
        //********************************************************************************
        private IntPtr InitializeDCForWmCtlColor(IntPtr dc, int msg) {
            IntPtr  brushRetVal;

            if (!GetStyle(ControlStyles.UserPaint)) {
                User32.SetTextColor(dc, ColorTranslator.ToWin32(ForeColor));
                User32.SetBkColor(dc, ColorTranslator.ToWin32(BackColor));
                brushRetVal = BackBrush;
            } else {
                brushRetVal = User32.GetStockObject(5);
            }
            return(brushRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// InvalidateEverything:       Redraw the control
        /// </summary>
        //
        //********************************************************************************
        private void InvalidateEverything() {
            User32.RedrawWindow(Handle, null, new HandleRef(null, IntPtr.Zero), 0x485);
        }

        //********************************************************************************
        //
        /// <summary>
        /// IsInputKey:                 Checks if the key is used by this control
        /// </summary>
		/// <returns>
        /// Handle of the window with the focus
        /// </returns>
        //
        //********************************************************************************
        protected override bool IsInputKey(Keys keyData) {
            bool    bRetVal;
            Keys    key;

            key = keyData & (Keys.Alt | Keys.KeyCode);
            if ((key == Keys.Return || key == Keys.Escape) && DroppedDown) {
                bRetVal = true;
            } else {
                if ((keyData & Keys.Alt) == Keys.Alt) {
                    bRetVal = false;
                } else {
                    switch(keyData & Keys.KeyCode) {
                    case Keys.Up:
                    case Keys.Down:
                    case Keys.Prior:
                    case Keys.Next:
                    case Keys.End:
                    case Keys.Home:
                        bRetVal = true;
                        break;
                    default:
                        bRetVal = base.IsInputKey(keyData);
                        break;
                    }
                }
            }
            return(bRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnBackColorChanged:         Called when background color changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnBackColorChanged(EventArgs e) {
            base.OnBackColorChanged(e);
            UpdateControl(false);
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnDropDown:                 Called when the drop down state change
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnDropDown(EventArgs e) {
            if (DropDown != null) {
                DropDown(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnDropDownStyleChanged:     Called when the style of the combo change
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnDropDownStyleChanged(EventArgs e) {
            if (DropDownStyleChanged != null) {
                DropDownStyleChanged(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnFontChanged:              Called when the font is changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnFontChanged(EventArgs e) {
            base.OnFontChanged(e);
            UpdateControl(true);
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnForeColorChanged:         Called when foreground color is changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnForeColorChanged(EventArgs e) {
            base.OnForeColorChanged(e);
            UpdateControl(false);
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnGotFocus:                 Called when control get the focus
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected override void OnGotFocus(EventArgs e) {
            if (!m_bCanFireLostFocus) {
                base.OnGotFocus(e);
                if (AutoDropDown) {
                    User32.PostMessage(Handle, (int)VCBM.SHOWDROPDOWN, 1, 0);
                }
                m_bCanFireLostFocus = true;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnHandleCreated:            Called when window handle is created
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnHandleCreated(EventArgs e) {
            int             iMaxLength;
            int             iDropDownWidth;
            int             iSelectedIndex;
            int             iItemCount;
            IntPtr          wnd;
            WndProc         proc;
            ComboBoxStyle   eStyle;

            base.OnHandleCreated(e);
            iMaxLength      = MaxLength;
            iSelectedIndex  = m_iSelectedIndex;
            iDropDownWidth  = DropDownWidth;
            eStyle          = DropDownStyle;
            if (iMaxLength > 0) {
                SendMessage(VCBM.LIMITTEXT, iMaxLength, 0);
            }

            if (!m_childWindowRoot.IsAllocated      &&
                m_wndListBox    == IntPtr.Zero      && 
                m_procListBox   == IntPtr.Zero      &&
                m_wndEdit       == IntPtr.Zero      && 
                m_procEdit      == IntPtr.Zero      &&
                DropDownStyle   != ComboBoxStyle.DropDownList) {
                wnd     = User32.GetWindow(Handle, GW.CHILD);
                if (wnd != IntPtr.Zero) {
                    proc              = new WndProc(new VComboBox.ChildWindow(this).Callback);
                    m_childWindowRoot = GCHandle.Alloc(proc);
                    if (DropDownStyle == ComboBoxStyle.Simple) {
                        m_wndListBox    = wnd;
                        m_procListBox   = User32.GetWindowLong(wnd, GWL.WNDPROC);
                        User32.SetWindowLong(wnd, GWL.WNDPROC, proc);
                        wnd             = User32.GetWindow(Handle, GW.HWNDNEXT);
                    }
                    m_wndEdit   = wnd;
                    m_procEdit  = User32.GetWindowLong(wnd, GWL.WNDPROC);
                    User32.SetWindowLong(wnd, GWL.WNDPROC, proc);
                }
            }
            if (iDropDownWidth >= 0) {
                SendMessage(VCBM.SETDROPPEDWIDTH, iDropDownWidth, 0);
            }
            if (this.DropDownStyle == ComboBoxStyle.Simple) {
                base.Height = m_iRequestedHeight;
            }
            iItemCount = m_iItemCount;
            SendMessage(VCBM.SETCOUNT, iItemCount, 0);
            if (iSelectedIndex >= 0) {
                SendMessage(VCBM.SETCURSEL, iSelectedIndex, 0);
                SendMessage(VCBM.SETTOPINDEX, iSelectedIndex, 0);
                UpdateText();
                m_iSelectedIndex = -1;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnHandleDestroyed:          Called when background color changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnHandleDestroyed(EventArgs e) {
            ReleaseChildWindow();
            m_wndDropDown = IntPtr.Zero;
            if (base.Disposing) {
                m_iSelectedIndex = -1;
                m_iItemCount     = 0;
            } else {
                m_iSelectedIndex = SelectedIndex;
                m_iItemCount     = ItemCount;
            }
            base.OnHandleDestroyed(e);
        }

        //********************************************************************************
        //
        /// <summary>
        /// ProcessAutoCompleteChar:    Process the auto-completion
        /// </summary>
        //
        //********************************************************************************
        protected virtual void ProcessAutoCompleteChar() {
            string          strCurText;
            int             iIndex;
            int             iPos;
            ItemTextArg     e;

            strCurText = Text;
            e          = new ItemTextArg(strCurText);
            OnFindItemIndex(e);
            iIndex     = e.Index;
            if (iIndex != -1) {
                if (SelectedIndex == iIndex) {
                    Text = GetItemTextFromIndex(iIndex);
                } else {
                    SelectedIndex = iIndex;
                }
                iPos = strCurText.Length;
                Select(iPos, Text.Length - iPos);
            } else {
                SelectedIndex = -1;
                Text          = strCurText;
                Select(Text.Length, 0);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnTextChanged:              Called when text edit text changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnTextChanged(EventArgs e) {
            base.OnTextChanged(e);
            if (m_bProcessAutoComplete) {
                m_bProcessAutoComplete = false;
                ProcessAutoCompleteChar();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnKeyPress:                 Called when a key is pressed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnKeyDown(KeyEventArgs e) {
            if (AutoComplete && DropDownStyle != ComboBoxStyle.DropDownList) {
                if (e.KeyCode == Keys.Delete) {
                    m_bProcessAutoComplete = true;
                }
            }
            base.OnKeyDown(e);
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnKeyPress:                 Called when a key is pressed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnKeyPress(KeyPressEventArgs e) {
            base.OnKeyPress(e);
            if (!e.Handled && (e.KeyChar == '\r' || e.KeyChar == '\x001b' || e.KeyChar == '\x09') && DroppedDown) {
                DroppedDown = false;
                e.Handled = true;
            }
            if (AutoComplete && DropDownStyle != ComboBoxStyle.DropDownList) {
                if (!Char.IsControl(e.KeyChar)) {
                    m_bProcessAutoComplete = true;
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnLostFocus:                Called when focus is lost
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        [EditorBrowsable(EditorBrowsableState.Advanced)]
        protected override void OnLostFocus(EventArgs e) {
            if (m_bCanFireLostFocus) {
                base.OnLostFocus(e);
                m_bCanFireLostFocus = false;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnParentBackColorChanged:   Called when parent background color is changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnParentBackColorChanged(EventArgs e) {
            base.OnParentBackColorChanged(e);
            if (DropDownStyle == ComboBoxStyle.Simple) {
                Invalidate();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnResize:                   Called when control is resized
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected override void OnResize(EventArgs e) {
            base.OnResize(e);
            if (DropDownStyle == ComboBoxStyle.Simple) {
                InvalidateEverything();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnSelectedIndexChanged:     Called when the selection is changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnSelectedIndexChanged(EventArgs e) {
            if (SelectedIndexChanged != null) {
                SelectedIndexChanged(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnSelectedItemChanged:      Called when the selected item is changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnSelectedItemChanged(EventArgs e) {
            if (SelectedItemChanged != null) {
                SelectedItemChanged(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnSelectedValueChanged:     Called when the selected value is changed
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnSelectedValueChanged(EventArgs e) {
            if (SelectedValueChanged != null) {
                SelectedValueChanged(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnSelectionChangeCommitted: Called when selection change is commited
        /// </summary>
        /// <param name="e">            Event argument</param>
        //
        //********************************************************************************
        protected virtual void OnSelectionChangeCommitted(EventArgs e) {
            if (SelectionChangeCommitted != null) {
                SelectionChangeCommitted(this, e);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// RefreshItems:               Refresh the items displayed in the list
        /// </summary>
        //
        //********************************************************************************
        private void RefreshItems() {
            //TODO
        }

        //********************************************************************************
        //
        /// <summary>
        /// ReleaseChildWindow:         Release the child window
        /// </summary>
        //
        //********************************************************************************
        private void ReleaseChildWindow() {
            if (m_wndEdit != IntPtr.Zero) {
                User32.SetWindowLong(m_wndEdit, GWL.WNDPROC, m_procEdit);
                m_wndEdit   = IntPtr.Zero;
                m_procEdit  = IntPtr.Zero;
            }
            if (m_wndListBox != IntPtr.Zero) {
                User32.SetWindowLong(m_wndListBox, GWL.WNDPROC, m_procListBox);
                m_wndListBox    = IntPtr.Zero;
                m_procListBox   = IntPtr.Zero;
            }
            if (m_childWindowRoot.IsAllocated) {
                m_childWindowRoot.Free();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Select:                     Select the text in the edit box
        /// </summary>
        /// <param name="iStart">       Start of the selection</param>
        /// <param name="iLength">      Length of the selection</param>
        //
        //********************************************************************************
        public void Select(int iStart, int iLength) {
            int iEnd;
            int iParam;

            if (iStart < 0) {
                throw new ArgumentException("Invalid argument - Start " + iStart.ToString());
            }
            if (iLength < 0) {
                throw new ArgumentException("Invalid argument - Length " + iLength.ToString());
            }
            iEnd    = iStart + iLength;
            iParam  = iStart + (iEnd << 16);
            SendMessage(VCBM.SETEDITSEL, 0, iParam);
        }

        //********************************************************************************
        //
        /// <summary>
        /// SelectAll:                  Select all the text in the edit box
        /// </summary>
        //
        //********************************************************************************
        public void SelectAll() {
            Select(0, 0x7fffffff);
        }

        //********************************************************************************
        //
        /// <summary>
        /// SetBoundsCore:              Performs the work of setting the specified bounds of this control. 
        /// </summary>
        /// <param name="x">            New Left of the control</param>
        /// <param name="y">            New Top of the control</param>
        /// <param name="width">        New Width of the control</param>
        /// <param name="height">       New Height of the control</param>
        /// <param name="specified">    A bitwise combination of the BoundsSpecified values</param>
        //
        //********************************************************************************
        protected override void SetBoundsCore(int x, int y, int width, int height, BoundsSpecified specified) {
            if ((specified & BoundsSpecified.Height) != BoundsSpecified.None) {
                m_iRequestedHeight = height;
            }
            if ((DropDownStyle == ComboBoxStyle.DropDown || DropDownStyle == ComboBoxStyle.DropDownList) && Parent != null) {
                height = PreferredHeight;
            }
            base.SetBoundsCore(x, y, width, height, specified);
        }

        //********************************************************************************
        //
        /// <summary>
        /// ToString:                   Convert the object to a string
        /// </summary>
        /// <returns>
        /// String
        /// </returns>
        //
        //********************************************************************************
        public override string ToString() {
            string strRetVal;
            
            strRetVal = base.ToString();
            return (strRetVal + ", Items.Count: " + ItemCount.ToString());
        }

        //********************************************************************************
        //
        /// <summary>
        /// UpdateControl:              Updates the control
        /// </summary>
        /// <param name="bRecreate">    true to recreate</param>
        //
        //********************************************************************************
        private void UpdateControl(bool bRecreate) {
            m_iPreferredHeight = -1;
            if (IsHandleCreated) {
                if ((DropDownStyle == ComboBoxStyle.Simple) && bRecreate) {
                    RecreateHandle();
                } else {
                    InvalidateEverything();
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// UpdateDropDownHeight:           Update the size of the drop down
        /// </summary>
        //
        //********************************************************************************
        private void UpdateDropDownHeight() {
            int     iItemCount;
            int     iMaxDropDownItems;

            if (m_wndDropDown != IntPtr.Zero) {
                iItemCount          = ItemCount;
                iMaxDropDownItems   = Math.Min(Math.Max(iItemCount, 1), MaxDropDownItems);
                User32.SetWindowPos(m_wndDropDown, IntPtr.Zero, 0, 0, DropDownWidth, ItemHeight * iMaxDropDownItems + 2, SWP.NOMOVE | SWP.NOZORDER);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// UpdateText:                 Updates the edit text
        /// </summary>
        //
        //********************************************************************************
        private void UpdateText() {
            string          strText = null;
            int             iSelectedIndex;
            
            iSelectedIndex = SelectedIndex;
            if (iSelectedIndex != -1) {
                strText = GetItemTextFromIndex(iSelectedIndex);
            }
            Text = strText;
            if (DropDownStyle == ComboBoxStyle.DropDown && m_wndEdit != IntPtr.Zero) {
                User32.SendMessageW(m_wndEdit, (int)WM.SETTEXT, 0, strText);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// WmEraseBkgnd:               Erase the background
        /// </summary>
        /// <param name="m">            Message structure</param>
        //
        //********************************************************************************
        private void WmEraseBkgnd(ref Message m) {
            RECT        tRect;
            Control     ctlParent;
            Graphics    gr;
            Brush       brush;

            if (DropDownStyle == ComboBoxStyle.Simple && Parent != null) {
                tRect = new RECT();
                User32.GetClientRect(Handle, ref tRect);
                ctlParent   = Parent;
                gr          = Graphics.FromHdc(m.WParam);
                if (ctlParent != null) {
                    brush = new SolidBrush(ctlParent.BackColor);
                    gr.FillRectangle(brush, tRect.left, tRect.top, tRect.right - tRect.left, tRect.bottom - tRect.top);
                    brush.Dispose();
                } else {
                  gr.FillRectangle(SystemBrushes.Control, tRect.left, tRect.top, tRect.right - tRect.left, tRect.bottom - tRect.top);
                }
                gr.Dispose();
                m.Result = (IntPtr)1;
            } else {
                base.WndProc(ref m);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// WmParentNotify:             Notify the parent
        /// </summary>
        /// <param name="m">            Message structure</param>
        //
        //********************************************************************************
        private void WmParentNotify(ref Message m) {
            base.WndProc(ref m);
            if (((int)m.WParam) == 0x3e80001) {
                m_wndDropDown = m.LParam;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// WmReflectCommand:               Reflect the WM_COMMAND message
        /// </summary>
        /// <param name="m">                Message</param>
        //
        //********************************************************************************
        private void WmReflectCommand(ref Message m) {
            switch ((VCBCMDN)(((int) m.WParam) >> 0x10)) {
            case VCBCMDN.SELCHANGE:
                UpdateText();
                OnSelectedIndexChanged(EventArgs.Empty);
                return;
            case VCBCMDN.DBLCLK:
            case VCBCMDN.SETFOCUS:
            case VCBCMDN.KILLFOCUS:
            case VCBCMDN.EDITUPDATE:
            case VCBCMDN.CLOSEUP:
                return;
            case VCBCMDN.EDITCHANGE:
                OnTextChanged(EventArgs.Empty);
                return;
            case VCBCMDN.DROPDOWN:
                OnDropDown(EventArgs.Empty);
                UpdateDropDownHeight();
                return;
            //case CBN_SELENDOK:
            //    OnSelectionChangeCommitted(EventArgs.Empty);
            //    return;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// WmReflectNotify:                Reflect the WM_NOTIFY message
        /// </summary>
        /// <param name="m">                Message</param>
        //
        //********************************************************************************
        private void WmReflectNotify(ref Message m) {
            VIRTUALCOMBOBOXITEM item;
            byte[]              byteArr;
            char[]              charArr;
            string              strItemText;

            item = (VIRTUALCOMBOBOXITEM)m.GetLParam(typeof(VIRTUALCOMBOBOXITEM));
            switch((VCBN)item.code) {
            case VCBN.GETDISPINFO:
                strItemText = GetItemTextFromIndex(item.iIndex) + "\0";
                if (Marshal.SystemDefaultCharSize == 1) {
                    byteArr = Encoding.Default.GetBytes(strItemText);
                    Marshal.Copy(byteArr, 0, item.lpszText, Math.Min(byteArr.Length, 255));
                } else {
                    charArr = strItemText.ToCharArray();
                    Marshal.Copy(charArr, 0, item.lpszText, Math.Min(charArr.Length, 255));
                }
                break;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// WndProc:                        Window procedure
        /// </summary>
        /// <param name="m">                Message</param>
        //
        //********************************************************************************
        protected override void WndProc(ref Message m) {
            int         iMsgID = m.Msg;
            RECT        tRect;
            Rectangle   rect;
            Point       pt;

            switch((WM)iMsgID) {
            case WM.SETFOCUS:
                try {
                    m_bFireSetFocus = false;
                    base.WndProc(ref m);
                } finally {
                    m_bFireSetFocus = true;
                }
                return;
            case WM.KILLFOCUS:
                try {
                    m_bFireLostFocus = false;
                    base.WndProc(ref m);
                } finally {
                    m_bFireLostFocus = true;
                }
                return;
            case WM.ERASEBKGND:
                WmEraseBkgnd(ref m);
                return;
            case WM.DRAWITEM:
            case WM.MEASUREITEM:
            case WM.DELETEITEM:
            case WM.COMPAREITEM:
                DefWndProc(ref m);
                return;
            case WM.CTLCOLOREDIT:
            case WM.CTLCOLORLISTBOX:
                m.Result = InitializeDCForWmCtlColor(m.WParam, m.Msg);
                return;
            case WM.LBUTTONDOWN:
                m_bMouseEvents = true;
                DefWndProc(ref m);
                return;
            case WM.LBUTTONUP:
                tRect   = new RECT();
                User32.GetWindowRect(Handle, ref tRect);
                rect    = new Rectangle(tRect.left, tRect.top, tRect.right - tRect.left, tRect.bottom - tRect.top);
                pt      = new Point((int)(short)(int)m.LParam, ((int)m.LParam) >> 0x10);
                pt      = PointToScreen(pt);
                if (m_bMouseEvents /* && !base.ValidationCancelled*/) {
                    m_bMouseEvents = false;
                    if (Capture && rect.Contains(pt)) {
                        OnClick(EventArgs.Empty);
                    }
                    base.WndProc(ref m);
                    return;
                }
                Capture = false;
                DefWndProc(ref m);
                return;
            case WM.PARENTNOTIFY:
                WmParentNotify(ref m);
                return;
            case WM.MOUSELEAVE:
                DefWndProc(ref m);
                tRect = new RECT();
                User32.GetWindowRect(Handle, ref tRect);
                rect  = new Rectangle(tRect.left, tRect.top, tRect.right - tRect.left, tRect.bottom - tRect.top);
                pt    = MousePosition;
                if (!rect.Contains(pt)) {
                    OnMouseLeave(EventArgs.Empty);
                    m_bMouseInEdit = false;
                }
                return;
            case (WM)((int)WM.REFLECT + (int)WM.COMMAND):
                  this.WmReflectCommand(ref m);
                  return;
            case (WM)((int)WM.REFLECT + (int)WM.NOTIFY):
                  this.WmReflectNotify(ref m);
                  return;
            }
            if (m.Msg == User32.WM_MOUSEENTER) {
                DefWndProc(ref m);
                if (m_bMouseInEdit) {
                    return;
                }
                OnMouseEnter(EventArgs.Empty);
                m_bMouseInEdit = true;
            } else {
                base.WndProc(ref m);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// BackgroundImage:            Background image
        /// </summary>
        //
        //********************************************************************************
        [EditorBrowsable(EditorBrowsableState.Never), Browsable(false)]
        public override Image BackgroundImage {
            get {
                return(base.BackgroundImage);
            }
            set {
                base.BackgroundImage = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// CreateParams:                Parameter used to create the window
        /// </summary>
        //
        //********************************************************************************
        protected override CreateParams CreateParams {
            get {
                CreateParams param;

                param                   = base.CreateParams;
                param.ClassName         = "VCOMBOBOXW";
                //param.Style             |= 0;//(int)(/*WS.VSCROLL | */WS.BORDER);
                //param.ExStyle            |= (int)WS_EX.CLIENTEDGE;
                switch(DropDownStyle) {
                case ComboBoxStyle.Simple:
                    param.Style |= (int)CBS.SIMPLE;
                    break;
                case ComboBoxStyle.DropDown:
                    param.Style |= (int)CBS.DROPDOWN;
                    param.Height = PreferredHeight;
                    break;
                case ComboBoxStyle.DropDownList:
                    param.Style |= (int)CBS.DROPDOWNLIST;
                    param.Height = PreferredHeight;
                    break;
                }
                return(param);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// DefaultSize:                Default size of the control
        /// </summary>
        //
        //********************************************************************************
        protected override Size DefaultSize {
            get {
                return new Size(0x79, PreferredHeight);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// DropDownStyle:          Controls the appearance and functionality of the combo box
        /// </summary>
        //
        //********************************************************************************
        [Description("Controls the appearance and functionality of the combo box.")]
        [Category("Appearance")]
        [RefreshProperties(RefreshProperties.Repaint)]
        [DefaultValue(1)]
        public ComboBoxStyle DropDownStyle {
            get {
                return(m_eComboStyle);
            }
            set {
                if (m_eComboStyle != value) {
                    if (!Enum.IsDefined(typeof(ComboBoxStyle), value)) {
                        throw new InvalidEnumArgumentException("value", (int) value, typeof(ComboBoxStyle));
                    }
                    m_eComboStyle = value;
                    if (base.IsHandleCreated) {
                        base.RecreateHandle();
                    }
                    OnDropDownStyleChanged(EventArgs.Empty);
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// DropDownWidth:              The width, in pixels, of the drop-down in a combo box
        /// </summary>
        //
        //********************************************************************************
        [Description("The width, in pixels, of the drop-down in a combo box.")]
        [Category("Behavior")]
        public int DropDownWidth {
            get {
                return((m_iDropDownWidth == -1) ? base.Width : m_iDropDownWidth);
            }
            set {
                if (value < 1) {
                    throw new ArgumentException("Invalid Argument - " + value.ToString());
                }
                if (m_iDropDownWidth != value) {
                    m_iDropDownWidth = value;
                    if (base.IsHandleCreated) {
                        SendMessage(VCBM.SETDROPPEDWIDTH, value, 0);
                    }
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// DropDown:               State of the drop-down
        /// </summary>
        //
        //********************************************************************************
        [Browsable(false)]
        [Description("State of the drop-down.")]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public bool DroppedDown {
            get {
                bool    bRetVal;

                if (IsHandleCreated) {
                    bRetVal = (((int)SendMessage(VCBM.GETDROPPEDSTATE, 0, 0)) != 0);
                } else {
                    bRetVal = false;
                }
                return(bRetVal);
            }
            set {
                SendMessage(VCBM.SHOWDROPDOWN, value ? -1 : 0, 0);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Focused:                Returns if the control is having the focus
        /// </summary>
        //
        //********************************************************************************
        public override bool Focused {
            get {
                bool    bRetVal;
                IntPtr  wndFocus;

                bRetVal = base.Focused;
                if (!bRetVal) {
                    wndFocus = User32.GetFocus();
                    if (wndFocus == IntPtr.Zero) {
                        bRetVal = false;
                    } else {
                        bRetVal = (wndFocus == m_wndEdit ||wndFocus == m_wndListBox);
                    }
                }
                return(bRetVal);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// IntegralHeight:             Gets or sets a value indicating whether the control should resize to avoid showing partial items.
        /// </summary>
        //
        //********************************************************************************
        [Category("Behavior")]
        [Description("Gets or sets a value indicating whether the control should resize to avoid showing partial items.")]
        [DefaultValue(true)]
        [Localizable(true)]
        public bool IntegralHeight {
            get {
                return(m_bIntegralHeight);
            }
            set {
                if (m_bIntegralHeight != value) {
                    m_bIntegralHeight = value;
                    RecreateHandle();
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ItemCount:                  Number of items in the combo box.
        /// </summary>
        //
        //********************************************************************************
        [Description("Number of items in the combo box.")]
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public virtual int ItemCount {
            get {
                int     iRetVal;

                if (base.IsHandleCreated) {
                    iRetVal = (int)SendMessage(VCBM.GETCOUNT, 0, 0);
                } else {
                    iRetVal = m_iItemCount;
                }
                return(iRetVal);
            }
            set {
                if (ItemCount != value) {
                    if (value < 0) {
                        throw new ArgumentOutOfRangeException("InvalidArgument index - " + value.ToString());
                    }
                    if (IsHandleCreated) {
                        SendMessage(VCBM.SETCOUNT, value, 0);
                    } else {
                        m_iItemCount = value;
                    }
                }
                UpdateText();
                if (IsHandleCreated) {
                    OnTextChanged(EventArgs.Empty);
                }
                OnSelectedItemChanged(EventArgs.Empty);
                OnSelectedIndexChanged(EventArgs.Empty);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ItemHeight:             Gets the height of an item in the combo box.
        /// </summary>
        //
        //********************************************************************************
        [Localizable(true)]
        [Description("Gets the height of an item in the combo box.")]
        [Category("Behavior")]
        public int ItemHeight {
            get {
                return(FontHeight + 2);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// MaxDropDownItems:       The maximum number of entries to display in the drop-down list
        /// </summary>
        //
        //********************************************************************************
        [Description("The maximum number of entries to display in the drop-down list.")]
        [Localizable(true)]
        [Category("Behavior")]
        [DefaultValue(8)]
        public int MaxDropDownItems {
            get {
                return(m_iMaxDropDownItems);
            }
            set {
                if (value < 1 || value > 100) {
                    throw new ArgumentOutOfRangeException("Invalid bound - " + value.ToString() + " (1-100)");
                }
                m_iMaxDropDownItems = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// MaxLength:          Specifies the maximum number of characters that can be entered into the combo box
        /// </summary>
        //
        //********************************************************************************
        [Category("Behavior")]
        [Localizable(true)]
        [Description("Specifies the maximum number of characters that can be entered into the combo box.")]
        [DefaultValue(0)]
        public int MaxLength {
            get {
                return(m_iMaxLength);
            }
            set {
                if (value < 0) {
                    value = 0;
                }
                if (MaxLength != value) {
                    m_iMaxLength = value;
                    if (IsHandleCreated) {
                        SendMessage(VCBM.LIMITTEXT, value, 0);
                    }
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// PreferredHeight:        Preferred height of the combo box
        /// </summary>
        //
        //********************************************************************************
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("Preferred height of the combo box.")]
        public int PreferredHeight {
            get {
                if (m_iPreferredHeight < 0) {
                    m_iPreferredHeight = FontHeight + ((SystemInformation.BorderSize.Height * 4) + 3);
                }
                return(m_iPreferredHeight);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SelectedIndex:          Currently selected item in the combo box.
        /// </summary>
        //
        //********************************************************************************
        [Description("Currently selected item in the combo box.")]
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public virtual int SelectedIndex {
            get {
                int     iRetVal;

                if (base.IsHandleCreated) {
                    iRetVal = (int)SendMessage(VCBM.GETCURSEL, 0, 0);
                } else {
                    iRetVal = m_iSelectedIndex;
                }
                return(iRetVal);
            }
            set {
                int     iCount;

                if (SelectedIndex != value) {
                    iCount = ItemCount;
                    if (value < -1 || value >= iCount) {
                        throw new ArgumentOutOfRangeException("Invalid argument - index " + value.ToString());
                    }
                    if (IsHandleCreated) {
                        SendMessage(VCBM.SETCURSEL, value, 0);
                        SendMessage(VCBM.SETTOPINDEX, value, 0);
                    } else {
                        m_iSelectedIndex = value;
                    }
                    UpdateText();
                    if (IsHandleCreated) {
                        OnTextChanged(EventArgs.Empty);
                    }
                    OnSelectedItemChanged(EventArgs.Empty);
                    OnSelectedIndexChanged(EventArgs.Empty);
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SelectedText:           Gets or sets the text that is selected in the editable portion of a ComboBox.
        /// </summary>
        //
        //********************************************************************************
        [Description("Gets or sets the text that is selected in the editable portion of a ComboBox.")]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Browsable(false)]
        public string SelectedText {
            get {
                string  strRetVal;

                if (DropDownStyle == ComboBoxStyle.DropDownList) {
                    strRetVal = "";
                } else {
                    strRetVal = Text.Substring(this.SelectionStart, this.SelectionLength);
                }
                return(strRetVal);
            }
            set {
                if (DropDownStyle != ComboBoxStyle.DropDownList) {
                    CreateControl();
                    SendMessage(VCBM.SETEDITSELTEXT, 0, value);
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SelectionLength:            Gets or sets the number of characters selected in the editable portion of the combo box.
        /// </summary>
        //
        //********************************************************************************
        [Description("Gets or sets the number of characters selected in the editable portion of the combo box.")]
        [Browsable(false)]
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public int SelectionLength {
            get {
                int     iStart = 0;
                int     iEnd = 0;

                SendMessage(VCBM.GETEDITSEL, ref iStart, ref iEnd);
                return(iEnd - iStart);
            }
            set {
                if (value < 0) {
                    throw new ArgumentException("Invalid argument value - " + value.ToString());
                }
                Select(SelectionStart, value);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// SelectionStart:             Gets or sets the starting index of text selected in the combo box.
        /// </summary>
        //
        //********************************************************************************
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        [Description("Gets or sets the starting index of text selected in the combo box.")]
        [Browsable(false)]
        public int SelectionStart {
            get {
                int     iStart = 0;
                int     iEnd = 0;

                SendMessage(VCBM.GETEDITSEL, ref iStart, ref iEnd);
                return(iStart);
            }
            set {
                if (value < 0) {
                    throw new ArgumentException("Invalid argument value - " + value.ToString());
                }
                Select(value, SelectionLength);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Text:                           Gets or sets the text associated with this control.
        /// </summary>
        //
        //********************************************************************************
        [Bindable(true)]
        [Localizable(true)]
        public override string Text {
            get {
                return(base.Text);
            }
            set  {
                ItemTextArg     e;
                int             iIndex;
                string          strItemText;

                if (base.Text != value) {
                    base.Text = value;
                    if (!base.DesignMode) {
                        if (value == null) {
                            iIndex = -1;
                        } else {
                            e = new ItemTextArg(value);
                            OnFindItemIndex(e);
                            iIndex     = e.Index;
                            if (iIndex != -1) {
                                strItemText = GetItemTextFromIndex(iIndex);
                                if (strItemText != value) {
                                    iIndex = -1;
                                }
                            }
                        }
                        if (SelectedIndex != iIndex) {
                            SelectedIndex = iIndex;
                        }
                    }
                }
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// BackBrush:                  Background brush
        /// </summary>
        //
        //********************************************************************************
        private IntPtr BackBrush {
            get {
                IntPtr  brushRetVal;
                Color   colBack;

                if (m_brushBack != (IntPtr)0) {
                    brushRetVal = m_brushBack;
                } else {
                    colBack = BackColor;
                    if (ColorTranslator.ToOle(colBack) < 0) {
                        brushRetVal = User32.GetSysColorBrush(ColorTranslator.ToOle(colBack) & 0xff);
                    } else {
                        brushRetVal = User32.CreateSolidBrush(ColorTranslator.ToWin32(colBack));
                    }
                    m_brushBack = brushRetVal;
                }
                return(brushRetVal);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// AutoComplete:               Specifies the mode for the automatic completion feature used in the ComboBox control
        /// </summary>
        //
        //********************************************************************************
        [Description("Specifies the mode for the automatic completion feature used in the ComboBox control. ")]
        [Category("Behavior")]
        public virtual bool AutoComplete {
            get {
                return(m_bAutoComplete);
            }
            set {
                m_bAutoComplete = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// AutoDropDown:               Automatically show the drop down when the control has the focus
        /// </summary>
        //
        //********************************************************************************
        [Description("Automatically show the drop down when the control has the focus. ")]
        [Category("Behavior")]
        public virtual bool AutoDropDown {
            get {
                return(m_bAutoDropDown);
            }
            set {
                m_bAutoDropDown = value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ChildWindow:                Implement a reference on a child window
        /// </summary>
        //
        //********************************************************************************
        private class ChildWindow {
            private WeakReference   m_ref;

            //********************************************************************************
            //
            /// <summary>
            /// ctor:                       Class constructor
            /// </summary>
            /// <param name="comboBox">     Attached father window</param>
            //
            //********************************************************************************
            public ChildWindow(VComboBox comboBox) {
                m_ref = new WeakReference(comboBox, false);
            }

            //********************************************************************************
            //
            /// <summary>
            /// ctor:                       Class constructor
            /// </summary>
            /// <param name="hWnd">         Window handle</param>
            /// <param name="iMsg">         Message id</param>
            /// <param name="wParam">       First parameter</param>
            /// <param name="lParam">       Second parameter</param>
            /// <returns>
            /// Result
            /// </returns>
            //
            //********************************************************************************
            public IntPtr Callback(IntPtr hWnd, int iMsg, IntPtr wParam, IntPtr lParam) {
                Message msg = Message.Create(hWnd, iMsg, wParam, lParam);
                
                try {
                    ((VComboBox)m_ref.Target).ChildWndProc(ref msg);
                } catch (Exception ex) {
                    System.Windows.Forms.Application.OnThreadException(ex);
                }
                return(msg.Result);
            }
        } // class ChildWindow
    } // Class VComboBox
} // Namespace
